  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//
// GizmoDaemon.cpp
//
//	GizmoDaemon class
//
///////////////////////////////////////
  ///////////////////////////////////////
  
/*

Copyright (c) 2005-2006, Tim Burrell
All rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "GizmoDaemon.hpp"
#include "../libGizmo/Debug.hpp"
#include "../libGizmo/UtilTime.hpp"
#include <signal.h>

////////////////////////////////////////////////////////////////////////////
// Statics / Type Defs
///////////////////////////////////////

#define DEFAULT_SCRIPT_PLUGIN		"GizmoScriptPluginPython"
#define CONFIGURATION_SCRIPT		"gizmoDaemonConfig"

#define FUNC_SHUTTINGDOWN		"gizmoShuttingDown"
#define FUNC_GIZMOPRINT			"gizmoPrint"

GizmoDaemon * GizmoDaemon::pThis = NULL;

////////////////////////////////////////////////////////////////////////////
// Overrides
///////////////////////////////////////

void GizmoDaemon:: onScriptInitialized() {
	m_ScriptInitialized = true;
}

////////////////////////////////////////////////////////////////////////////
// Global Funcs
///////////////////////////////////////

void GizmoDaemon::gizmoDaemonSignalHandler(int Signal) {
	if (!GizmoDaemon::pThis) {
		Debug::outWarn("GizmoDaemon :: SignalHandler: Daemon Does Not Exist!");
		return;
	}
		
	GizmoDaemon::pThis->signalHandler(Signal);
}
	
////////////////////////////////////////////////////////////////////////////
// Construction
///////////////////////////////////////

GizmoDaemon::GizmoDaemon(int argc, char ** argv) : 
	CommandLineOptions(argc, argv), 
	m_ScriptFuncWaitForNextEvent(this),
	m_ScriptFuncLoadPlugin(&m_PluginList),
	m_ScriptFuncIsPluginLoaded(&m_PluginList),
	m_ScriptFuncIsPluginLoadedString(&m_PluginList)
{
	pThis = this;
	m_ScriptPluginName = DEFAULT_SCRIPT_PLUGIN;
	m_ConfigScript = CONFIGURATION_SCRIPT;
	m_InitError = false;
	m_pScriptEngine = NULL;
	m_Looping = true;
	m_ScriptInitialized = false;

	init(argc, argv);
}

GizmoDaemon::~GizmoDaemon() {
}

////////////////////////////////////////////////////////////////////////////
// Class Body
///////////////////////////////////////

void GizmoDaemon::signalHandler(int Signal) {
	static bool ProcessingSignal = false;
	static int  KeybInt = 0;
	if (ProcessingSignal)
		return;
	ProcessingSignal = true;

	switch (Signal) {
	case SIGSEGV:
		Debug::outError("Segmentation Fault! [%d]", Signal);
		m_Looping = false;
		break;
		
	case SIGINT:
		if (!GizmoDaemon::pThis->m_ScriptInitialized) {
			Debug::outInfo("Keyboard Interrupt Received... Ignored until Program Initialized!");
			break;
		}
		if (KeybInt > 0)
			return;
		KeybInt ++;
		Debug::outInfo("Keyboard Interrupt Received...");
		m_Looping = false;
		break;

	case SIGHUP:
		if (!GizmoDaemon::pThis->m_ScriptInitialized) {
			Debug::outInfo("Interrupt Received [HUP]... Ignored until Program Initialized!");
			break;
		}
		Debug::outInfo("Carrier Loss Signal Received [%d]", Signal);
		m_Looping = false;
		break;

	case SIGQUIT:
		if (!GizmoDaemon::pThis->m_ScriptInitialized) {
			Debug::outInfo("Interrupt Received [QUIT]... Ignored until Program Initialized!");
			break;
		}
		Debug::outInfo("Core Dump Request Signal Received [%d]", Signal);
		m_Looping = false;
		break;

	case SIGKILL:
		if (!GizmoDaemon::pThis->m_ScriptInitialized) {
			Debug::outInfo("Interrupt Received [KILL]... Ignored until Program Initialized!");
			break;
		}
		Debug::outInfo("Unconditional Termination Signal Received [%d]", Signal);
		m_Looping = false;
		break;

	case SIGTERM:
		if (!GizmoDaemon::pThis->m_ScriptInitialized) {
			Debug::outInfo("Interrupt Received [TERM]... Ignored until Program Initialized!");
			break;
		}
		Debug::outInfo("Termination Signal Received [%d]", Signal);
		m_Looping = false;
		break;

	case SIGSTOP:
		if (!GizmoDaemon::pThis->m_ScriptInitialized) {
			Debug::outInfo("Interrupt Received [STOP]... Ignored until Program Initialized!");
			break;
		}
		Debug::outInfo("Unconditional Stop Signal Received [%d]", Signal);
		m_Looping = false;
		break;
		
	default:
		Debug::outInfo("Unknown Signal Received [%d]", Signal);
	}

	signal(Signal, &gizmoDaemonSignalHandler);

	ProcessingSignal = false;
}

int GizmoDaemon::enterLoop() {	
	if (m_InitError) {
		Debug::outDebug("Not Entering Processing Loop Due to Init Error!");
		return -1;
	}
					
	Debug::outDebug("GizmoDaemon :: Processing Events...");

	// run the configuration script
	if (m_pScriptEngine->runScript(m_ConfigScript)) {
		Debug::outError("Failed to execute config script [%s]", CONFIGURATION_SCRIPT);
		return false;
	}
	
	Debug::out("Shutting Down...");
	
	return 0;
}

void GizmoDaemon::displayProps() {
	Debug::out("\n%s v%s -=- (c) 2006, %s", PACKAGE_NAME, PACKAGE_VERSION, PACKAGE_BUGREPORT);
	Debug::out("-----------------\n");
}

void GizmoDaemon::displayUsage() {
	displayProps();
	
	printf("Usage: gizmod <arguments>\n\n");
	printf("    -g --debug              :: Enable debug mode\n");
	printf("    -h --help               :: You're looking at it\n");
	printf("    -q --quiet              :: Enable quiet mode\n");
	printf("    -s --script-plugin      :: Override Default Script Plugin\n");
	printf("                               - Default is '%s'\n", DEFAULT_SCRIPT_PLUGIN);
	printf("    -c --config-script      :: Override Default Script File\n");
	printf("                               - Default is '%s'\n", CONFIGURATION_SCRIPT);
	printf("    -v --version            :: Display Version Information\n");
}

bool GizmoDaemon::init(int argc, char ** argv) {
	m_InitError = true;

	// catch signals
	signal(SIGINT,   gizmoDaemonSignalHandler);
	signal(SIGHUP,   gizmoDaemonSignalHandler);
	signal(SIGQUIT,  gizmoDaemonSignalHandler);
	signal(SIGKILL,  gizmoDaemonSignalHandler);
	signal(SIGTERM,  gizmoDaemonSignalHandler);
	signal(SIGSTOP,  gizmoDaemonSignalHandler);
	
	// set up command line options
	addOptionRule('g', "debug", false);
	addOptionRule('h', "help", false);
	addOptionRule('q', "quiet", false);
	addOptionRule('s', "script-plugin", true);
	addOptionRule('c', "config-script", true);
	addOptionRule('v', "version", false);
	
	processCommandLine();

	for (int lp = 0; lp < getNumOptions(); lp ++) {
		CLOptionArg * pOpt = getOption(lp);

		if (pOpt->error) {
			printf("Invalid Argument Specified.\n\n");
			displayUsage();
			return true;
		}

		switch (pOpt->shortOption) {
		case 'g':
			Debug::setDebugMode(true);
			break;
		default:
			printf("Invalid Argument Specified.\n\n");
		case 'h':
			displayUsage();
			return true;
		case 'q':
			Debug::setDisableAllOutput(true);
			break;
		case 's':
			m_ScriptPluginName = pOpt->argument;
			break;
		case 'c':
			m_ConfigScript = pOpt->argument;
			break;
		case 'v':
			Debug::out("\nVersion %s [%s]", PACKAGE_VERSION, PACKAGE_WHIMSICAL_NAME);
			return true;
		}
	}

	displayProps();
	
	// load the scripting plugin
	if (!(loadScriptPlugin())) {
		Debug::outError("Failed to Load Script Plugin [%s]!", m_ScriptPluginName.getString());
		return false;
	}
		
	// made it this far, no error!
	m_InitError = false;

	return true;
}

bool GizmoDaemon::loadScriptPlugin() {
	Debug::outDebug("Loading Gizmo Scripting Plugin [%s]...", m_ScriptPluginName.getString());

	// load the plugin
	if ((m_pScriptEngine = m_ScriptPlugin.loadPlugin(m_ScriptPluginName, NULL)) == NULL)
		return false;
	
	// register with LoadPlugin
	m_ScriptFuncLoadPlugin.setScriptEngine(m_pScriptEngine);
		
	// register the functions
	if (!m_pScriptEngine->registerScriptFunction(&m_ScriptFuncGetVersion))
		return false;
	if (!m_pScriptEngine->registerScriptFunction(&m_ScriptFuncWaitForNextEvent))
		return false;
	if (!m_pScriptEngine->registerScriptFunction(&m_ScriptFuncLoadPlugin))
		return false;
	if (!m_pScriptEngine->registerScriptFunction(&m_ScriptFuncIsPluginLoaded))
		return false;
	if (!m_pScriptEngine->registerScriptFunction(&m_ScriptFuncIsPluginLoadedString))
		return false;
			
	// call plugin's init routine
	if (!m_pScriptEngine->initPlugin()) {
		Debug::outError("Failed to Load Gizmo Scripting Plugin [%s]", m_ScriptPluginName.getString());
		return false;
	}

	//Debug::out("Loaded Gizmo Scripting Plugin [%s]", m_ScriptPluginName.getString());
	
	return true;
}

bool GizmoDaemon::getInitError() {
	return m_InitError;
}

bool GizmoDaemon::signalPluginsToCheckForEvents() {
	m_ScriptFuncLoadPlugin.checkPluginsForPendingEvents();
	return !m_Looping;
}
