////////////////////////////////////////////////////////////////////////////
//
// CommandLineOptions.cpp
//
//	CommandLineOptions main file
//
///////////////////////////////////////

/*

Copyright (c) 2005, Tim Burrell
All rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "CommandLineOptions.hpp"
#include "Debug.hpp"

////////////////////////////////////////////////////////////////////////////
// Statics / Type Defs
///////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// Construction
///////////////////////////////////////

CommandLineOptions::CommandLineOptions(int argc, char * argv []) {
	createOptionList(argc, argv);
}

CommandLineOptions::~CommandLineOptions() {
	deleteArgList();
	deleteOptionList();
	deleteRuleList();
}

////////////////////////////////////////////////////////////////////////////
// Overridables
///////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// Class Body
///////////////////////////////////////

bool CommandLineOptions::createOptionList(int argc, char * argv []) {
	deleteOptionList();

	for (int lp = 1; lp < argc; lp ++) {
		if (!argv[lp])
			continue;

		CLOptionNode * pNode = new CLOptionNode;
		pNode->arg = argv[lp];

		// check for short options compacted together
		if ( (pNode->arg.beginsWith("-")) && (!pNode->arg.beginsWith("--")) && (pNode->arg.getLength() > 2) ) {
			// short option
			for (int ilp = 1; ilp < pNode->arg.getLength(); ilp ++) {
				CLOptionNode * ptNode = new CLOptionNode;
				ptNode->arg = "-" + pNode->arg.mid(ilp, ilp);
				m_OptionList.addTail(ptNode);
			}

			// clean up
			pNode->arg.freeString();
			delete pNode;
		} else
			m_OptionList.addTail(pNode);
	}

	return true;
}

void CommandLineOptions::deleteOptionList() {
	if (m_OptionList.getSize() == 0)
		return;

	m_OptionList.moveToHead();
	do {
		CLOptionNode * pNode = m_OptionList.getData();
		pNode->arg.freeString();
		delete pNode;
	} while (m_OptionList.moveNext());

	m_OptionList.deleteAll();
}

void CommandLineOptions::deleteRuleList() {
	if (m_RuleList.getSize() == 0)
		return;

	m_RuleList.moveToHead();
	do {
		CLOptionRule * pRule = m_RuleList.getData();
		pRule->longOption.freeString();
		delete pRule;
	} while (m_RuleList.moveNext());

	m_RuleList.deleteAll();
}

void CommandLineOptions::deleteArgList() {
	if (m_ArgList.getSize() == 0)
		return;

	m_ArgList.moveToHead();
	do {
		CLOptionArg * pArg = m_ArgList.getData();
		pArg->argument.freeString();
		delete pArg;
	} while (m_ArgList.moveNext());

	m_ArgList.deleteAll();
}

void CommandLineOptions::addOptionRule(char shortOption, String longOption, bool hasArgument) {
	CLOptionRule * pRule = new CLOptionRule;
	pRule->shortOption = shortOption;
	pRule->longOption = longOption;
	pRule->hasArgument = hasArgument;

	m_RuleList.addTail(pRule);
}

bool CommandLineOptions::processCommandLine() {
	if (m_RuleList.getSize() == 0)
		return false;
		
	deleteArgList();
	
	if (m_OptionList.getSize() == 0)
		return true;

	m_OptionList.moveToHead();
	do {
		CLOptionNode * pNode = m_OptionList.getData();
		if (!pNode)
			continue;

		CLOptionArg * 	pArg = NULL;
		String 		Arg;
		bool 		isShort;
		
		// check if it's a long or a short, or neither (say a filename or whatever)
		if (pNode->arg.beginsWith("--")) {
			// long option
			Arg=pNode->arg.mid(2);
			isShort = false;
		} else if (pNode->arg.beginsWith("-")) {
			// short option
			Arg=pNode->arg.mid(1);
			if (Arg.getLength() > 1) {
				// short option is long!
				isShort = false;
			} else
				isShort = true;
		} else {
			// neither			
			pArg = new CLOptionArg;
			pArg->shortOption = CLOPTION_STANDALONE;
			pArg->argument = pNode->arg;
			pArg->error = false;
			
			m_ArgList.addTail(pArg);
			continue;
		}

		// so now we just have to find if it's in the rule list
		bool Found = false;
		m_RuleList.moveToHead();
		do {
			CLOptionRule * pRule = m_RuleList.getData();
			if (!pRule)
				continue;
							
			bool Match = false;
			if ( (isShort) && (pRule->shortOption == Arg.getCharAt(0)) ) {
				// we have a short match
				Match = true;
			} else if (pRule->longOption == Arg) {
				// we have a long match
				Match = true;
			}

			if (!Match)
				// no Match!
				continue;

			// check if there's an argument
			String	Argument;
			bool	Error = false;
			if (pRule->hasArgument) {
				if (!m_OptionList.moveNext()) {
					// no argument!
					Error = true;
				} else {
					CLOptionNode * tpNode = m_OptionList.getData();
					if (tpNode->arg.beginsWith("-")) {
						// no argument!
						m_OptionList.movePrev();
						Error = true;
					} else
						Argument = tpNode->arg;
				}
			}

			// we've gotten to this point, so there must be a match!
			CLOptionArg * pArg = new CLOptionArg;
			pArg->shortOption = pRule->shortOption;
			pArg->argument = Argument;
			pArg->error = Error;
			m_ArgList.addTail(pArg);
			m_RuleList.moveToTail();
			Found = true;
		} while (m_RuleList.moveNext());

		// if the option was not found at all (ie invalid) report it
		if (!Found) {
			CLOptionArg * pArg = new CLOptionArg;
			pArg->shortOption = CLOPTION_ERROR;
			pArg->argument = Arg;
			pArg->error = true;
			m_ArgList.addTail(pArg);
		}		
	} while (m_OptionList.moveNext());

	return true;
}

int CommandLineOptions::getNumOptions() {
	return m_ArgList.getSize();
}

CLOptionArg * CommandLineOptions::getOption(int index) {
	if (!m_ArgList.moveTo(index))
		return NULL;

	return m_ArgList.getData();
}

