  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//
// GizmoDeviceInput.cpp
//
//	GizmoDeviceInput class
//
///////////////////////////////////////
  ///////////////////////////////////////
  
/*

Copyright (c) 2005, Tim Burrell
All rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "GizmoDeviceInput.hpp"
#include "libGizmo/Debug.hpp"
#include "libGizmo/UtilTime.hpp"
#include <linux/input.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>

////////////////////////////////////////////////////////////////////////////
// Fixes for buggy distros
///////////////////////////////////////

#ifndef KEY_ZENKAKUHANKAKU
	#define KEY_ZENKAKUHANKAKU	85
#endif
#ifndef KEY_RO
	#define KEY_RO			89
#endif
#ifndef KEY_KATAKANA
	#define KEY_KATAKANA		90
#endif
#ifndef KEY_HIRAGANA
	#define KEY_HIRAGANA		91
#endif
#ifndef KEY_HENKAN
	#define KEY_HENKAN		92
#endif
#ifndef KEY_KATAKANAHIRAGANA
	#define KEY_KATAKANAHIRAGANA	93
#endif
#ifndef KEY_MUHENKAN
	#define KEY_MUHENKAN		94
#endif
#ifndef KEY_KPJPCOMMA
	#define KEY_KPJPCOMMA		95
#endif
#ifndef KEY_HANGUEL
	#define KEY_HANGUEL		122
#endif
#ifndef KEY_HANJA
	#define KEY_HANJA		123
#endif
#ifndef KEY_YEN
	#define KEY_YEN			124
#endif

////////////////////////////////////////////////////////////////////////////
// Statics / Type Defs
///////////////////////////////////////

#define DEVICE_OPENMODE		O_RDWR
#define NUM_EVENT_DEVICES	16
#define IBUF_SIZE		32

int GizmoDeviceInput::m_DevTotal = 0;

////////////////////////////////////////////////////////////////////////////
// Overrides
///////////////////////////////////////

void GizmoDeviceInput::onInputSyn(int Code, int Value) {
	//Debug::outDebug("GizmoDeviceInput :: onInputSyn [%d] [%d]", Code, Value);
}

void GizmoDeviceInput::onInputKey(int Code, int Value) {
	//Debug::outDebug("GizmoDeviceInput :: onInputKey [%d] [%d]", Code, Value);
}

void GizmoDeviceInput::onInputRelative(int Code, int Value) {
	//Debug::outDebug("GizmoDeviceInput :: onInputRelative [%d] [%d]", Code, Value);
}

void GizmoDeviceInput::onInputAbsolute(int Code, int Value) {
	//Debug::outDebug("GizmoDeviceInput :: onInputAbsolute [%d] [%d]", Code, Value);
}

void GizmoDeviceInput::onInputMisc(int Code, int Value) {
	//Debug::outDebug("GizmoDeviceInput :: onInputMisc [%d] [%d]", Code, Value);
}

void GizmoDeviceInput::onInputLED(int Code, int Value) {
	//Debug::outDebug("GizmoDeviceInput :: onInputLED [%d] [%d]", Code, Value);
}
	
////////////////////////////////////////////////////////////////////////////
// Construction
///////////////////////////////////////

GizmoDeviceInput::GizmoDeviceInput(const char * DevicePath) {
	m_DevNum = m_DevTotal ++;
	
	m_DevicePath = DevicePath;
	m_FD = -1;
}

GizmoDeviceInput::~GizmoDeviceInput() {
	closeDevice();
}

////////////////////////////////////////////////////////////////////////////
// Class Body
///////////////////////////////////////

void GizmoDeviceInput::closeDevice() {
	if (m_FD != -1)
		close(m_FD);

	m_FD = -1;
}

bool GizmoDeviceInput::initDevice() {
	if (!openDevice(m_DevicePath)) {
		Debug::outDebug("GizmoDeviceInput :: Event Input Device Failed to Initialize!");
		return false;
	}
	
	return true;
}

bool GizmoDeviceInput::openDevice(const char * DevName, int DeviceNumber) {
	int curnum = 0;
	char tDevName[256] = {'\0'};
	char CheckName[256] = {'\0'};
	if (m_FD != -1)
		close(m_FD);
	// try opening as a /dev path first
	int fd = open(DevName, DEVICE_OPENMODE);
	if (fd < 0) {
		// not a valid dev, try each input device by name
		for (int i = 0; i < NUM_EVENT_DEVICES; i ++) {
			sprintf(tDevName, "/dev/input/event%d", i);
			if ((fd = open(tDevName, DEVICE_OPENMODE)) >= 0) {
				if (ioctl(fd, EVIOCGNAME(sizeof(CheckName)), CheckName) < 0) {
					Debug::outError("GizmoDeviceInput :: [%s] : EVIOCGNAME failed: %s", CheckName, strerror(errno));
					close(fd);
					continue;
				}
				if (String(DevName) == "!ListDevices") {
					Debug::out("\t- Device [%s] : [%s]", tDevName, CheckName);
					continue;
				}
				Debug::outDebug("Comparing [%s] and [%s]", CheckName, DevName);
				if (strcasestr(CheckName, DevName) != NULL) {
					if (DeviceNumber >= 0) {
						curnum ++;
						if (curnum != DeviceNumber)
							continue;
					}
					m_FD = fd;
					m_DevicePath = tDevName;
					m_DeviceName = CheckName;
					Debug::outDebug("GizmoDeviceInput :: Event Device [%s] found at [%s]", CheckName, m_DevicePath.getString());
					return true;
				}
			}
		}
	} else {
		if (ioctl(fd, EVIOCGNAME(sizeof(CheckName)), CheckName) < 0) {
			Debug::outError("GizmoDeviceInput :: [%s] : EVIOCGNAME failed: %s", CheckName, strerror(errno));
			closeDevice();
			return false;
		}
		m_FD = fd;
		m_DevicePath = DevName;
		m_DeviceName = CheckName;
		Debug::outDebug("GizmoDeviceInput :: Event Device [%s] found at [%s]", DevName, m_DevicePath.getString());
		return true;
	}

	if (String(DevName) == "!ListDevices")
		return true;

	// if we got here, we've failed
	Debug::outDebug("GizmoDeviceInput :: Unable to open \"%s\"\n    - Check Perms in /dev/input!", DevName);
	return false;
}

/*
 * Adapted from Sowerbutt's Event user land utils
 * Thanks to all your work in getting this stuff to work!
 * -- Modified to use non blocking IO
 */
bool GizmoDeviceInput::checkEventsWaiting() {
	if (m_FD == -1)
		return false;
		
	struct input_event ibuffer[IBUF_SIZE];
	int 		r, events, i;
	fd_set 		readfds;
	struct timeval	TimeOut;
	bool		ret = false;

	TimeOut.tv_sec = 0;
	TimeOut.tv_usec = 0;
		
	// tell select which sockets to poll
	FD_ZERO(&readfds);
	FD_SET(m_FD, &readfds);

	// wait on sockets
	int NumSelected = select(m_FD + 1, &readfds, NULL, NULL, &TimeOut);
	if (NumSelected == -1)
		return false;
	else if (NumSelected == 0) // timeout
		return false;
			
	if (FD_ISSET(m_FD, &readfds)) {
		// read event!
		r = read(m_FD, ibuffer, sizeof(struct input_event) * IBUF_SIZE);
		if (r > 0) {
			ret = true;
			events = r / sizeof(struct input_event);
			for (i = 0; i < events; i ++)
				processEvent(&ibuffer[i]);
		} else {
			Debug::outError("GizmoDeviceInput :: Failed to Read from Device!: %s", strerror(errno));
		}
	}
	
	return ret;
}
	
void GizmoDeviceInput::processEvent(struct input_event *ev) {
	//Debug::out("InputDevice%s :: [type=0x%04x code=0x%04x, value=%d]", m_DeviceName.getString(), ev->type, ev->code, (int)ev->value);
	switch(ev->type) {
	case EV_SYN:
		onInputSyn(ev->code, ev->value);
		break;
	case EV_REL:
		onInputRelative(ev->code, ev->value);
		break;
	case EV_KEY:
		onInputKey(ev->code, ev->value);
		break;
	case EV_ABS:
		onInputAbsolute(ev->code, ev->value);
		break;
	case EV_MSC:
		onInputMisc(ev->code, ev->value);
		break;
	case EV_LED:
		onInputLED(ev->code, ev->value);
		break;
	default:
		Debug::outDebug("InputDevice%s :: [type=0x%04x code=0x%04x, value=%d]", m_DeviceName.getString(), ev->type, ev->code, (int)ev->value);
	}
}

bool GizmoDeviceInput::writeToDevice(int Type, int Code, int Value) {
	//Debug::out("GizmoDeviceInput :: Writing [%d, %d, %d] to [%s]", Type, Code, Value, m_DeviceName.getString());
	struct input_event ev = {0};
	ev.type = Type;
	ev.code = Code;
	ev.value = Value;
	if (write(m_FD, &ev, sizeof(struct input_event)) == -1)
		return false;
	return true;
}

bool GizmoDeviceInput::remapKey(int CurCode, int NewCode) {
	int codes[2];
	codes[0] = CurCode;
	codes[1] = NewCode;
	if (ioctl(m_FD, EVIOCSKEYCODE, codes)) {
		Debug::outDebug("GizmoDeviceInput :: Key Remapping Failed!");
		return false;
	}

	return true;
}

bool GizmoDeviceInput::grabExclusiveAccess(bool Grab) {
	if (ioctl(m_FD, EVIOCGRAB, Grab ? 1 : 0)) {
		Debug::outDebug("GizmoDeviceInput :: Device Grabbing Failed!");
		return false;
	}

	return true;
}

const char * GizmoDeviceInput::getKeyStringFromCode(int Code) {
	switch (Code) {
	case KEY_RESERVED:
		return "KEY_RESERVED";
	case KEY_ESC:
		return "KEY_ESC";
	case KEY_1:
		return "KEY_1";
	case KEY_2:
		return "KEY_2";
	case KEY_3:
		return "KEY_3";
	case KEY_4:
		return "KEY_4";
	case KEY_5:
		return "KEY_5";
	case KEY_6:
		return "KEY_6";
	case KEY_7:
		return "KEY_7";
	case KEY_8:
		return "KEY_8";
	case KEY_9:
		return "KEY_9";
	case KEY_0:
		return "KEY_0";
	case KEY_MINUS:
		return "KEY_MINUS";
	case KEY_EQUAL:
		return "KEY_EQUAL";
	case KEY_BACKSPACE:
		return "KEY_BACKSPACE";
	case KEY_TAB:
		return "KEY_TAB";
	case KEY_Q:
		return "KEY_Q";
	case KEY_W:
		return "KEY_W";
	case KEY_E:
		return "KEY_E";
	case KEY_R:
		return "KEY_R";
	case KEY_T:
		return "KEY_T";
	case KEY_Y:
		return "KEY_Y";
	case KEY_U:
		return "KEY_U";
	case KEY_I:
		return "KEY_I";
	case KEY_O:
		return "KEY_O";
	case KEY_P:
		return "KEY_P";
	case KEY_LEFTBRACE:
		return "KEY_LEFTBRACE";
	case KEY_RIGHTBRACE:
		return "KEY_RIGHTBRACE";
	case KEY_ENTER:
		return "KEY_ENTER";
	case KEY_LEFTCTRL:
		return "KEY_LEFTCTRL";
	case KEY_A:
		return "KEY_A";
	case KEY_S:
		return "KEY_S";
	case KEY_D:
		return "KEY_D";
	case KEY_F:
		return "KEY_F";
	case KEY_G:
		return "KEY_G";
	case KEY_H:
		return "KEY_H";
	case KEY_J:
		return "KEY_J";
	case KEY_K:
		return "KEY_K";
	case KEY_L:
		return "KEY_L";
	case KEY_SEMICOLON:
		return "KEY_SEMICOLON";
	case KEY_APOSTROPHE:
		return "KEY_APOSTROPHE";
	case KEY_GRAVE:
		return "KEY_GRAVE";
	case KEY_LEFTSHIFT:
		return "KEY_LEFTSHIFT";
	case KEY_BACKSLASH:
		return "KEY_BACKSLASH";
	case KEY_Z:
		return "KEY_Z";
	case KEY_X:
		return "KEY_X";
	case KEY_C:
		return "KEY_C";
	case KEY_V:
		return "KEY_V";
	case KEY_B:
		return "KEY_B";
	case KEY_N:
		return "KEY_N";
	case KEY_M:
		return "KEY_M";
	case KEY_COMMA:
		return "KEY_COMMA";
	case KEY_DOT:
		return "KEY_DOT";
	case KEY_SLASH:
		return "KEY_SLASH";
	case KEY_RIGHTSHIFT:
		return "KEY_RIGHTSHIFT";
	case KEY_KPASTERISK:
		return "KEY_KPASTERISK";
	case KEY_LEFTALT:
		return "KEY_LEFTALT";
	case KEY_SPACE:
		return "KEY_SPACE";
	case KEY_CAPSLOCK:
		return "KEY_CAPSLOCK";
	case KEY_F1:
		return "KEY_F1";
	case KEY_F2:
		return "KEY_F2";
	case KEY_F3:
		return "KEY_F3";
	case KEY_F4:
		return "KEY_F4";
	case KEY_F5:
		return "KEY_F5";
	case KEY_F6:
		return "KEY_F6";
	case KEY_F7:
		return "KEY_F7";
	case KEY_F8:
		return "KEY_F8";
	case KEY_F9:
		return "KEY_F9";
	case KEY_F10:
		return "KEY_F10";
	case KEY_NUMLOCK:
		return "KEY_NUMLOCK";
	case KEY_SCROLLLOCK:
		return "KEY_SCROLLLOCK";
	case KEY_KP7:
		return "KEY_KP7";
	case KEY_KP8:
		return "KEY_KP8";
	case KEY_KP9:
		return "KEY_KP9";
	case KEY_KPMINUS:
		return "KEY_KPMINUS";
	case KEY_KP4:
		return "KEY_KP4";
	case KEY_KP5:
		return "KEY_KP5";
	case KEY_KP6:
		return "KEY_KP6";
	case KEY_KPPLUS:
		return "KEY_KPPLUS";
	case KEY_KP1:
		return "KEY_KP1";
	case KEY_KP2:
		return "KEY_KP2";
	case KEY_KP3:
		return "KEY_KP3";
	case KEY_KP0:
		return "KEY_KP0";
	case KEY_KPDOT:
		return "KEY_KPDOT";
	case KEY_ZENKAKUHANKAKU:
		return "KEY_ZENKAKUHANKAKU";
	case KEY_102ND:
		return "KEY_102ND";
	case KEY_F11:
		return "KEY_F11";
	case KEY_F12:
		return "KEY_F12";
	case KEY_RO:
		return "KEY_RO";
	case KEY_KATAKANA:
		return "KEY_KATAKANA";
	case KEY_HIRAGANA:
		return "KEY_HIRAGANA";
	case KEY_HENKAN:
		return "KEY_HENKAN";
	case KEY_KATAKANAHIRAGANA:
		return "KEY_KATAKANAHIRAGANA";
	case KEY_MUHENKAN:
		return "KEY_MUHENKAN";
	case KEY_KPJPCOMMA:
		return "KEY_KPJPCOMMA";
	case KEY_KPENTER:
		return "KEY_KPENTER";
	case KEY_RIGHTCTRL:
		return "KEY_RIGHTCTRL";
	case KEY_KPSLASH:
		return "KEY_KPSLASH";
	case KEY_SYSRQ:
		return "KEY_SYSRQ";
	case KEY_RIGHTALT:
		return "KEY_RIGHTALT";
	case KEY_LINEFEED:
		return "KEY_LINEFEED";
	case KEY_HOME:
		return "KEY_HOME";
	case KEY_UP:
		return "KEY_UP";
	case KEY_PAGEUP:
		return "KEY_PAGEUP";
	case KEY_LEFT:
		return "KEY_LEFT";
	case KEY_RIGHT:
		return "KEY_RIGHT";
	case KEY_END:
		return "KEY_END";
	case KEY_DOWN:
		return "KEY_DOWN";
	case KEY_PAGEDOWN:
		return "KEY_PAGEDOWN";
	case KEY_INSERT:
		return "KEY_INSERT";
	case KEY_DELETE:
		return "KEY_DELETE";
	case KEY_MACRO:
		return "KEY_MACRO";
	case KEY_MUTE:
		return "KEY_MUTE";
	case KEY_VOLUMEDOWN:
		return "KEY_VOLUMEDOWN";
	case KEY_VOLUMEUP:
		return "KEY_VOLUMEUP";
	case KEY_POWER:
		return "KEY_POWER";
	case KEY_KPEQUAL:
		return "KEY_KPEQUAL";
	case KEY_KPPLUSMINUS:
		return "KEY_KPPLUSMINUS";
	case KEY_PAUSE:
		return "KEY_PAUSE";
	case KEY_KPCOMMA:
		return "KEY_KPCOMMA";
	case KEY_HANGUEL:
		return "KEY_HANGUEL";
	case KEY_HANJA:
		return "KEY_HANJA";
	case KEY_YEN:
		return "KEY_YEN";
	case KEY_LEFTMETA:
		return "KEY_LEFTMETA";
	case KEY_RIGHTMETA:
		return "KEY_RIGHTMETA";
	case KEY_COMPOSE:
		return "KEY_COMPOSE";
	case KEY_STOP:
		return "KEY_STOP";
	case KEY_AGAIN:
		return "KEY_AGAIN";
	case KEY_PROPS:
		return "KEY_PROPS";
	case KEY_UNDO:
		return "KEY_UNDO";
	case KEY_FRONT:
		return "KEY_FRONT";
	case KEY_COPY:
		return "KEY_COPY";
	case KEY_OPEN:
		return "KEY_OPEN";
	case KEY_PASTE:
		return "KEY_PASTE";
	case KEY_FIND:
		return "KEY_FIND";
	case KEY_CUT:
		return "KEY_CUT";
	case KEY_HELP:
		return "KEY_HELP";
	case KEY_MENU:
		return "KEY_MENU";
	case KEY_CALC:
		return "KEY_CALC";
	case KEY_SETUP:
		return "KEY_SETUP";
	case KEY_SLEEP:
		return "KEY_SLEEP";
	case KEY_WAKEUP:
		return "KEY_WAKEUP";
	case KEY_FILE:
		return "KEY_FILE";
	case KEY_SENDFILE:
		return "KEY_SENDFILE";
	case KEY_DELETEFILE:
		return "KEY_DELETEFILE";
	case KEY_XFER:
		return "KEY_XFER";
	case KEY_PROG1:
		return "KEY_PROG1";
	case KEY_PROG2:
		return "KEY_PROG2";
	case KEY_WWW:
		return "KEY_WWW";
	case KEY_MSDOS:
		return "KEY_MSDOS";
	case KEY_COFFEE:
		return "KEY_COFFEE";
	case KEY_DIRECTION:
		return "KEY_DIRECTION";
	case KEY_CYCLEWINDOWS:
		return "KEY_CYCLEWINDOWS";
	case KEY_MAIL:
		return "KEY_MAIL";
	case KEY_BOOKMARKS:
		return "KEY_BOOKMARKS";
	case KEY_COMPUTER:
		return "KEY_COMPUTER";
	case KEY_BACK:
		return "KEY_BACK";
	case KEY_FORWARD:
		return "KEY_FORWARD";
	case KEY_CLOSECD:
		return "KEY_CLOSECD";
	case KEY_EJECTCD:
		return "KEY_EJECTCD";
	case KEY_EJECTCLOSECD:
		return "KEY_EJECTCLOSECD";
	case KEY_NEXTSONG:
		return "KEY_NEXTSONG";
	case KEY_PLAYPAUSE:
		return "KEY_PLAYPAUSE";
	case KEY_PREVIOUSSONG:
		return "KEY_PREVIOUSSONG";
	case KEY_STOPCD:
		return "KEY_STOPCD";
	case KEY_RECORD:
		return "KEY_RECORD";
	case KEY_REWIND:
		return "KEY_REWIND";
	case KEY_PHONE:
		return "KEY_PHONE";
	case KEY_ISO:
		return "KEY_ISO";
	case KEY_CONFIG:
		return "KEY_CONFIG";
	case KEY_HOMEPAGE:
		return "KEY_HOMEPAGE";
	case KEY_REFRESH:
		return "KEY_REFRESH";
	case KEY_EXIT:
		return "KEY_EXIT";
	case KEY_MOVE:
		return "KEY_MOVE";
	case KEY_EDIT:
		return "KEY_EDIT";
	case KEY_SCROLLUP:
		return "KEY_SCROLLUP";
	case KEY_SCROLLDOWN:
		return "KEY_SCROLLDOWN";
	case KEY_KPLEFTPAREN:
		return "KEY_KPLEFTPAREN";
	case KEY_KPRIGHTPAREN:
		return "KEY_KPRIGHTPAREN";
	case KEY_F13:
		return "KEY_F13";
	case KEY_F14:
		return "KEY_F14";
	case KEY_F15:
		return "KEY_F15";
	case KEY_F16:
		return "KEY_F16";
	case KEY_F17:
		return "KEY_F17";
	case KEY_F18:
		return "KEY_F18";
	case KEY_F19:
		return "KEY_F19";
	case KEY_F20:
		return "KEY_F20";
	case KEY_F21:
		return "KEY_F21";
	case KEY_F22:
		return "KEY_F22";
	case KEY_F23:
		return "KEY_F23";
	case KEY_F24:
		return "KEY_F24";
	case KEY_PLAYCD:
		return "KEY_PLAYCD";
	case KEY_PAUSECD:
		return "KEY_PAUSECD";
	case KEY_PROG3:
		return "KEY_PROG3";
	case KEY_PROG4:
		return "KEY_PROG4";
	case KEY_SUSPEND:
		return "KEY_SUSPEND";
	case KEY_CLOSE:
		return "KEY_CLOSE";
	case KEY_PLAY:
		return "KEY_PLAY";
	case KEY_FASTFORWARD:
		return "KEY_FASTFORWARD";
	case KEY_BASSBOOST:
		return "KEY_BASSBOOST";
	case KEY_PRINT:
		return "KEY_PRINT";
	case KEY_HP:
		return "KEY_HP";
	case KEY_CAMERA:
		return "KEY_CAMERA";
	case KEY_SOUND:
		return "KEY_SOUND";
	case KEY_QUESTION:
		return "KEY_QUESTION";
	case KEY_EMAIL:
		return "KEY_EMAIL";
	case KEY_CHAT:
		return "KEY_CHAT";
	case KEY_SEARCH:
		return "KEY_SEARCH";
	case KEY_CONNECT:
		return "KEY_CONNECT";
	case KEY_FINANCE:
		return "KEY_FINANCE";
	case KEY_SPORT:
		return "KEY_SPORT";
	case KEY_SHOP:
		return "KEY_SHOP";
	case KEY_ALTERASE:
		return "KEY_ALTERASE";
	case KEY_CANCEL:
		return "KEY_CANCEL";
	case KEY_BRIGHTNESSDOWN:
		return "KEY_BRIGHTNESSDOWN";
	case KEY_BRIGHTNESSUP:
		return "KEY_BRIGHTNESSUP";
	case KEY_MEDIA:
		return "KEY_MEDIA";
	default:
	case KEY_UNKNOWN:
		return "KEY_UNKNOWN";
	}
}

int GizmoDeviceInput::getKeyCodeFromString(String inString) {
	if (inString == "KEY_RESERVED")
		return KEY_RESERVED;
	else if (inString == "KEY_ESC")
		return KEY_ESC;
	else if (inString == "KEY_1")
		return KEY_1;
	else if (inString == "KEY_2")
		return KEY_2;
	else if (inString == "KEY_3")
		return KEY_3;
	else if (inString == "KEY_4")
		return KEY_4;
	else if (inString == "KEY_5")
		return KEY_5;
	else if (inString == "KEY_6")
		return KEY_6;
	else if (inString == "KEY_7")
		return KEY_7;
	else if (inString == "KEY_8")
		return KEY_8;
	else if (inString == "KEY_9")
		return KEY_9;
	else if (inString == "KEY_0")
		return KEY_0;
	else if (inString == "KEY_MINUS")
		return KEY_MINUS;
	else if (inString == "KEY_EQUAL")
		return KEY_EQUAL;
	else if (inString == "KEY_BACKSPACE")
		return KEY_BACKSPACE;
	else if (inString == "KEY_TAB")
		return KEY_TAB;
	else if (inString == "KEY_Q")
		return KEY_Q;
	else if (inString == "KEY_W")
		return KEY_W;
	else if (inString == "KEY_E")
		return KEY_E;
	else if (inString == "KEY_R")
		return KEY_R;
	else if (inString == "KEY_T")
		return KEY_T;
	else if (inString == "KEY_Y")
		return KEY_Y;
	else if (inString == "KEY_U")
		return KEY_U;
	else if (inString == "KEY_I")
		return KEY_I;
	else if (inString == "KEY_O")
		return KEY_O;
	else if (inString == "KEY_P")
		return KEY_P;
	else if (inString == "KEY_LEFTBRACE")
		return KEY_LEFTBRACE;
	else if (inString == "KEY_RIGHTBRACE")
		return KEY_RIGHTBRACE;
	else if (inString == "KEY_ENTER")
		return KEY_ENTER;
	else if (inString == "KEY_LEFTCTRL")
		return KEY_LEFTCTRL;
	else if (inString == "KEY_A")
		return KEY_A;
	else if (inString == "KEY_S")
		return KEY_S;
	else if (inString == "KEY_D")
		return KEY_D;
	else if (inString == "KEY_F")
		return KEY_F;
	else if (inString == "KEY_G")
		return KEY_G;
	else if (inString == "KEY_H")
		return KEY_H;
	else if (inString == "KEY_J")
		return KEY_J;
	else if (inString == "KEY_K")
		return KEY_K;
	else if (inString == "KEY_L")
		return KEY_L;
	else if (inString == "KEY_SEMICOLON")
		return KEY_SEMICOLON;
	else if (inString == "KEY_APOSTROPHE")
		return KEY_APOSTROPHE;
	else if (inString == "KEY_GRAVE")
		return KEY_GRAVE;
	else if (inString == "KEY_LEFTSHIFT")
		return KEY_LEFTSHIFT;
	else if (inString == "KEY_BACKSLASH")
		return KEY_BACKSLASH;
	else if (inString == "KEY_Z")
		return KEY_Z;
	else if (inString == "KEY_X")
		return KEY_X;
	else if (inString == "KEY_C")
		return KEY_C;
	else if (inString == "KEY_V")
		return KEY_V;
	else if (inString == "KEY_B")
		return KEY_B;
	else if (inString == "KEY_N")
		return KEY_N;
	else if (inString == "KEY_M")
		return KEY_M;
	else if (inString == "KEY_COMMA")
		return KEY_COMMA;
	else if (inString == "KEY_DOT")
		return KEY_DOT;
	else if (inString == "KEY_SLASH")
		return KEY_SLASH;
	else if (inString == "KEY_RIGHTSHIFT")
		return KEY_RIGHTSHIFT;
	else if (inString == "KEY_KPASTERISK")
		return KEY_KPASTERISK;
	else if (inString == "KEY_LEFTALT")
		return KEY_LEFTALT;
	else if (inString == "KEY_SPACE")
		return KEY_SPACE;
	else if (inString == "KEY_CAPSLOCK")
		return KEY_CAPSLOCK;
	else if (inString == "KEY_F1")
		return KEY_F1;
	else if (inString == "KEY_F2")
		return KEY_F2;
	else if (inString == "KEY_F3")
		return KEY_F3;
	else if (inString == "KEY_F4")
		return KEY_F4;
	else if (inString == "KEY_F5")
		return KEY_F5;
	else if (inString == "KEY_F6")
		return KEY_F6;
	else if (inString == "KEY_F7")
		return KEY_F7;
	else if (inString == "KEY_F8")
		return KEY_F8;
	else if (inString == "KEY_F9")
		return KEY_F9;
	else if (inString == "KEY_F10")
		return KEY_F10;
	else if (inString == "KEY_NUMLOCK")
		return KEY_NUMLOCK;
	else if (inString == "KEY_SCROLLLOCK")
		return KEY_SCROLLLOCK;
	else if (inString == "KEY_KP7")
		return KEY_KP7;
	else if (inString == "KEY_KP8")
		return KEY_KP8;
	else if (inString == "KEY_KP9")
		return KEY_KP9;
	else if (inString == "KEY_KPMINUS")
		return KEY_KPMINUS;
	else if (inString == "KEY_KP4")
		return KEY_KP4;
	else if (inString == "KEY_KP5")
		return KEY_KP5;
	else if (inString == "KEY_KP6")
		return KEY_KP6;
	else if (inString == "KEY_KPPLUS")
		return KEY_KPPLUS;
	else if (inString == "KEY_KP1")
		return KEY_KP1;
	else if (inString == "KEY_KP2")
		return KEY_KP2;
	else if (inString == "KEY_KP3")
		return KEY_KP3;
	else if (inString == "KEY_KP0")
		return KEY_KP0;
	else if (inString == "KEY_KPDOT")
		return KEY_KPDOT;
	else if (inString == "KEY_ZENKAKUHANKAKU")
		return KEY_ZENKAKUHANKAKU;
	else if (inString == "KEY_102ND")
		return KEY_102ND;
	else if (inString == "KEY_F11")
		return KEY_F11;
	else if (inString == "KEY_F12")
		return KEY_F12;
	else if (inString == "KEY_RO")
		return KEY_RO;
	else if (inString == "KEY_KATAKANA")
		return KEY_KATAKANA;
	else if (inString == "KEY_HIRAGANA")
		return KEY_HIRAGANA;
	else if (inString == "KEY_HENKAN")
		return KEY_HENKAN;
	else if (inString == "KEY_KATAKANAHIRAGANA")
		return KEY_KATAKANAHIRAGANA;
	else if (inString == "KEY_MUHENKAN")
		return KEY_MUHENKAN;
	else if (inString == "KEY_KPJPCOMMA")
		return KEY_KPJPCOMMA;
	else if (inString == "KEY_KPENTER")
		return KEY_KPENTER;
	else if (inString == "KEY_RIGHTCTRL")
		return KEY_RIGHTCTRL;
	else if (inString == "KEY_KPSLASH")
		return KEY_KPSLASH;
	else if (inString == "KEY_SYSRQ")
		return KEY_SYSRQ;
	else if (inString == "KEY_RIGHTALT")
		return KEY_RIGHTALT;
	else if (inString == "KEY_LINEFEED")
		return KEY_LINEFEED;
	else if (inString == "KEY_HOME")
		return KEY_HOME;
	else if (inString == "KEY_UP")
		return KEY_UP;
	else if (inString == "KEY_PAGEUP")
		return KEY_PAGEUP;
	else if (inString == "KEY_LEFT")
		return KEY_LEFT;
	else if (inString == "KEY_RIGHT")
		return KEY_RIGHT;
	else if (inString == "KEY_END")
		return KEY_END;
	else if (inString == "KEY_DOWN")
		return KEY_DOWN;
	else if (inString == "KEY_PAGEDOWN")
		return KEY_PAGEDOWN;
	else if (inString == "KEY_INSERT")
		return KEY_INSERT;
	else if (inString == "KEY_DELETE")
		return KEY_DELETE;
	else if (inString == "KEY_MACRO")
		return KEY_MACRO;
	else if (inString == "KEY_MUTE")
		return KEY_MUTE;
	else if (inString == "KEY_VOLUMEDOWN")
		return KEY_VOLUMEDOWN;
	else if (inString == "KEY_VOLUMEUP")
		return KEY_VOLUMEUP;
	else if (inString == "KEY_POWER")
		return KEY_POWER;
	else if (inString == "KEY_KPEQUAL")
		return KEY_KPEQUAL;
	else if (inString == "KEY_KPPLUSMINUS")
		return KEY_KPPLUSMINUS;
	else if (inString == "KEY_PAUSE")
		return KEY_PAUSE;
	else if (inString == "KEY_KPCOMMA")
		return KEY_KPCOMMA;
	else if (inString == "KEY_HANGUEL")
		return KEY_HANGUEL;
	else if (inString == "KEY_HANJA")
		return KEY_HANJA;
	else if (inString == "KEY_YEN")
		return KEY_YEN;
	else if (inString == "KEY_LEFTMETA")
		return KEY_LEFTMETA;
	else if (inString == "KEY_RIGHTMETA")
		return KEY_RIGHTMETA;
	else if (inString == "KEY_COMPOSE")
		return KEY_COMPOSE;
	else if (inString == "KEY_STOP")
		return KEY_STOP;
	else if (inString == "KEY_AGAIN")
		return KEY_AGAIN;
	else if (inString == "KEY_PROPS")
		return KEY_PROPS;
	else if (inString == "KEY_UNDO")
		return KEY_UNDO;
	else if (inString == "KEY_FRONT")
		return KEY_FRONT;
	else if (inString == "KEY_COPY")
		return KEY_COPY;
	else if (inString == "KEY_OPEN")
		return KEY_OPEN;
	else if (inString == "KEY_PASTE")
		return KEY_PASTE;
	else if (inString == "KEY_FIND")
		return KEY_FIND;
	else if (inString == "KEY_CUT")
		return KEY_CUT;
	else if (inString == "KEY_HELP")
		return KEY_HELP;
	else if (inString == "KEY_MENU")
		return KEY_MENU;
	else if (inString == "KEY_CALC")
		return KEY_CALC;
	else if (inString == "KEY_SETUP")
		return KEY_SETUP;
	else if (inString == "KEY_SLEEP")
		return KEY_SLEEP;
	else if (inString == "KEY_WAKEUP")
		return KEY_WAKEUP;
	else if (inString == "KEY_FILE")
		return KEY_FILE;
	else if (inString == "KEY_SENDFILE")
		return KEY_SENDFILE;
	else if (inString == "KEY_DELETEFILE")
		return KEY_DELETEFILE;
	else if (inString == "KEY_XFER")
		return KEY_XFER;
	else if (inString == "KEY_PROG1")
		return KEY_PROG1;
	else if (inString == "KEY_PROG2")
		return KEY_PROG2;
	else if (inString == "KEY_WWW")
		return KEY_WWW;
	else if (inString == "KEY_MSDOS")
		return KEY_MSDOS;
	else if (inString == "KEY_COFFEE")
		return KEY_COFFEE;
	else if (inString == "KEY_DIRECTION")
		return KEY_DIRECTION;
	else if (inString == "KEY_CYCLEWINDOWS")
		return KEY_CYCLEWINDOWS;
	else if (inString == "KEY_MAIL")
		return KEY_MAIL;
	else if (inString == "KEY_BOOKMARKS")
		return KEY_BOOKMARKS;
	else if (inString == "KEY_COMPUTER")
		return KEY_COMPUTER;
	else if (inString == "KEY_BACK")
		return KEY_BACK;
	else if (inString == "KEY_FORWARD")
		return KEY_FORWARD;
	else if (inString == "KEY_CLOSECD")
		return KEY_CLOSECD;
	else if (inString == "KEY_EJECTCD")
		return KEY_EJECTCD;
	else if (inString == "KEY_EJECTCLOSECD")
		return KEY_EJECTCLOSECD;
	else if (inString == "KEY_NEXTSONG")
		return KEY_NEXTSONG;
	else if (inString == "KEY_PLAYPAUSE")
		return KEY_PLAYPAUSE;
	else if (inString == "KEY_PREVIOUSSONG")
		return KEY_PREVIOUSSONG;
	else if (inString == "KEY_STOPCD")
		return KEY_STOPCD;
	else if (inString == "KEY_RECORD")
		return KEY_RECORD;
	else if (inString == "KEY_REWIND")
		return KEY_REWIND;
	else if (inString == "KEY_PHONE")
		return KEY_PHONE;
	else if (inString == "KEY_ISO")
		return KEY_ISO;
	else if (inString == "KEY_CONFIG")
		return KEY_CONFIG;
	else if (inString == "KEY_HOMEPAGE")
		return KEY_HOMEPAGE;
	else if (inString == "KEY_REFRESH")
		return KEY_REFRESH;
	else if (inString == "KEY_EXIT")
		return KEY_EXIT;
	else if (inString == "KEY_MOVE")
		return KEY_MOVE;
	else if (inString == "KEY_EDIT")
		return KEY_EDIT;
	else if (inString == "KEY_SCROLLUP")
		return KEY_SCROLLUP;
	else if (inString == "KEY_SCROLLDOWN")
		return KEY_SCROLLDOWN;
	else if (inString == "KEY_KPLEFTPAREN")
		return KEY_KPLEFTPAREN;
	else if (inString == "KEY_KPRIGHTPAREN")
		return KEY_KPRIGHTPAREN;
	else if (inString == "KEY_F13")
		return KEY_F13;
	else if (inString == "KEY_F14")
		return KEY_F14;
	else if (inString == "KEY_F15")
		return KEY_F15;
	else if (inString == "KEY_F16")
		return KEY_F16;
	else if (inString == "KEY_F17")
		return KEY_F17;
	else if (inString == "KEY_F18")
		return KEY_F18;
	else if (inString == "KEY_F19")
		return KEY_F19;
	else if (inString == "KEY_F20")
		return KEY_F20;
	else if (inString == "KEY_F21")
		return KEY_F21;
	else if (inString == "KEY_F22")
		return KEY_F22;
	else if (inString == "KEY_F23")
		return KEY_F23;
	else if (inString == "KEY_F24")
		return KEY_F24;
	else if (inString == "KEY_PLAYCD")
		return KEY_PLAYCD;
	else if (inString == "KEY_PAUSECD")
		return KEY_PAUSECD;
	else if (inString == "KEY_PROG3")
		return KEY_PROG3;
	else if (inString == "KEY_PROG4")
		return KEY_PROG4;
	else if (inString == "KEY_SUSPEND")
		return KEY_SUSPEND;
	else if (inString == "KEY_CLOSE")
		return KEY_CLOSE;
	else if (inString == "KEY_PLAY")
		return KEY_PLAY;
	else if (inString == "KEY_FASTFORWARD")
		return KEY_FASTFORWARD;
	else if (inString == "KEY_BASSBOOST")
		return KEY_BASSBOOST;
	else if (inString == "KEY_PRINT")
		return KEY_PRINT;
	else if (inString == "KEY_HP")
		return KEY_HP;
	else if (inString == "KEY_CAMERA")
		return KEY_CAMERA;
	else if (inString == "KEY_SOUND")
		return KEY_SOUND;
	else if (inString == "KEY_QUESTION")
		return KEY_QUESTION;
	else if (inString == "KEY_EMAIL")
		return KEY_EMAIL;
	else if (inString == "KEY_CHAT")
		return KEY_CHAT;
	else if (inString == "KEY_SEARCH")
		return KEY_SEARCH;
	else if (inString == "KEY_CONNECT")
		return KEY_CONNECT;
	else if (inString == "KEY_FINANCE")
		return KEY_FINANCE;
	else if (inString == "KEY_SPORT")
		return KEY_SPORT;
	else if (inString == "KEY_SHOP")
		return KEY_SHOP;
	else if (inString == "KEY_ALTERASE")
		return KEY_ALTERASE;
	else if (inString == "KEY_CANCEL")
		return KEY_CANCEL;
	else if (inString == "KEY_BRIGHTNESSDOWN")
		return KEY_BRIGHTNESSDOWN;
	else if (inString == "KEY_BRIGHTNESSUP")
		return KEY_BRIGHTNESSUP;
	else if (inString == "KEY_MEDIA")
		return KEY_MEDIA;
	else
		return KEY_UNKNOWN;
}

