  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//
// PThread.cpp
//
//	PThread class
//
///////////////////////////////////////
  ///////////////////////////////////////
  
/*

Copyright (c) 2005, Tim Burrell
All rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "PThread.hpp"
#include "Debug.hpp"
#include "UtilTime.hpp"
#include <errno.h>

////////////////////////////////////////////////////////////////////////////
// Globals
///////////////////////////////////////

bool PThread::m_Stabilized = true;

void * PTHREAD_ThreadFunction(void * Arguments) {
	if (!Arguments)
		return NULL;

	void * ret;
	PThread * pPThread = static_cast<PThread *>(Arguments);
	if (!pPThread->m_ThreadFuncCallback)
		ret = pPThread->threadFunction();
	else
		ret = pPThread->m_ThreadFuncCallback(pPThread);
		
	pPThread->exitThread(ret);
	pPThread->onThreadExit(ret);
	pPThread->reset();
	
	return NULL;
}

void * PThread::threadFunction() {
	Debug::outError("PThread :: You are using PThread wrong.");

	return NULL;
}

////////////////////////////////////////////////////////////////////////////
// Type Defs
///////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// Construction
///////////////////////////////////////

PThread::PThread(void * Arguments) {
	m_ThreadID = 0;
	m_Threading = false;
	m_Arguments = Arguments;
	m_ExitSignalled = false;
	m_ThreadFuncCallback = NULL;
	m_ThreadExitFuncCallback = NULL;
	m_ClassName = "PThread";
	m_Waiting = false;
}

PThread::~PThread() {
	exitThread(NULL);
}

////////////////////////////////////////////////////////////////////////////
// Overridables
///////////////////////////////////////

void PThread::onThreadExit(void * ReturnVal) {
}

void PThread::onThreadExiting() {
}
	
////////////////////////////////////////////////////////////////////////////
// Class Body
///////////////////////////////////////

bool PThread::createThread() {
	if ( (m_ThreadID) || (m_ExitSignalled) )
		return true;
		
	m_Threading = true;
	
	int ret = pthread_create(&m_ThreadID, NULL, PTHREAD_ThreadFunction, this);

	if (ret != 0) {
		switch (ret) {
		case EAGAIN:
			Debug::outDebug("PThread :: Failed to Create Thread, Too Many Threads, or Not Enough Resources!");
			break;
		default:
			Debug::outDebug("PThread :: Failed to Create Thread, Unknown Error!");
		}
		reset();
		return false;
	}
	
	return true;
}

void PThread::reset() {
	m_ThreadID = 0;
	m_Threading = false;
	m_Arguments = NULL;
	m_ExitSignalled = false;
}

bool PThread::exitThread(void * ReturnVal) {
	m_ExitSignalled = true;

	if (m_ThreadExitFuncCallback)
		m_ThreadExitFuncCallback();
	else
		onThreadExiting();

	return true;
}

bool PThread::waitOnToExit(long MaxWait) {
	return waitOnThreadToExit(this, MaxWait);
}
		
bool PThread::waitOnThreadToExit(PThread * pPThread, long MaxWait) {
	if (!pPThread->m_Threading)
		return true;

	// CALL TWICE!?
	if (pPThread->m_Waiting) {
		Debug::outDebug("PThread :: waitOnThread Again!?");
		while (pPThread->m_Waiting) {
			// eat cpu
		}
	}

	pPThread->m_ExitSignalled = true;
		
	timespec waitTime;
	timespec retTime;
	waitTime.tv_sec = 0;
	waitTime.tv_nsec = 50000;

	long curWait = 0;
	while (pPThread->m_Threading) {
		pPThread->m_Waiting = true;
		if (!curWait)
			Debug::outDebug("PThread :: Waiting on Thread to Exit... %s", pPThread->getPThreadClassName());
		nanosleep(&waitTime, &retTime);
		curWait ++;
		if (MaxWait > 0) {
			if (curWait >= MaxWait) {
				Debug::outDebug("PThread :: Waited on Thread [%d] Units : %s", curWait, pPThread->getPThreadClassName());
				pPThread->m_Waiting = false;
				return false;
			}
		}
	}

	pPThread->m_Waiting = false;
	Debug::outDebug("PThread :: Waited on Thread [%d] Units : %s", curWait, pPThread->getPThreadClassName());
	
	return true;
}

void PThread::setPThreadCallback(void * (* ThreadFuncCallback) (PThread *)) {
	m_ThreadFuncCallback = ThreadFuncCallback;
}

void PThread::setPThreadExitCallback(void (* ThreadExitFuncCallback) (void)) {
	m_ThreadExitFuncCallback = ThreadExitFuncCallback;
}

bool PThread::hasExitSignalled() {
	return m_ExitSignalled;
}

void PThread::setPThreadName(const char * Name) {
	m_ClassName = Name;
}

const char * PThread::getPThreadClassName() {
	return m_ClassName;
}

bool PThread::isThreading() {
	return m_Threading;
}

void * PThread::getPThreadArguments() {
	return m_Arguments;
}

void PThread::setStabilized(bool Stable) {
	if (!(m_Stabilized = Stable))
		// wait to settle the system
		UtilTime::nanoSleep(250000000);
}

bool PThread::getStabilized() {
	return m_Stabilized;
}
