  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//
// SharedMemorySegment.cpp
//
//	SharedMemorySegment class
//
///////////////////////////////////////
  ///////////////////////////////////////
  
/*

Copyright (c) 2005, Tim Burrell
All rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "SharedMemorySegment.hpp"
#include "Debug.hpp"
#include "FileIO.hpp"

#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

////////////////////////////////////////////////////////////////////////////
// Globals, Defines, Type Defs
///////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// Overloads
///////////////////////////////////////

void SharedMemorySegment::onSharedMemorySegmentChange(void * data) {
	Debug::outDebug("SharedMemorySegment :: onChange");
}

////////////////////////////////////////////////////////////////////////////
// Construction
///////////////////////////////////////

SharedMemorySegment::SharedMemorySegment(int ID, unsigned int DataSize, bool HardClose) {
	m_Data = NULL;
	m_Delay = 1000;
	m_ID = ID;
	m_DataSize = DataSize;
	m_HardClose = HardClose;

	initSHM();
}

SharedMemorySegment::~SharedMemorySegment() {
	closeSHM();
}

////////////////////////////////////////////////////////////////////////////
// Class Body
///////////////////////////////////////

bool SharedMemorySegment::initSHM() {
	//if (!FileIO::touchRecursive(m_Path.getString()))
	//	Debug::outWarn("SharedMemorySegment :: Failed to Touch Shared Memory Segment [%s]", m_Path.getString());
	//m_SHMkey = ftok(m_Path.getString(), m_ID);
	//m_SHMid = shmget(m_SHMkey, m_DataSize, 0600 | IPC_CREAT);
	m_SHMid = shmget(m_ID, m_DataSize, 0600 | IPC_CREAT);
	if (m_SHMid == -1) {
		switch (errno) {
		case EACCES:
			Debug::outDebug("SharedMemorySegment :: Failed to Get Segment -- Access Denied");
			return false;
		case EEXIST:
			Debug::outDebug("SharedMemorySegment :: Failed to Get Segment -- Already Exists and Marked Exclusive!");
			return false;
		case EINVAL:
			Debug::outDebug("SharedMemorySegment :: Failed to Get Segment -- Already Exists and Size Incorrect!");
			return false;
		case ENFILE:
			Debug::outDebug("SharedMemorySegment :: Failed to Get Segment -- Too Many Files!");
			return false;
		case ENOENT:
			Debug::outDebug("SharedMemorySegment :: Failed to Get Segment -- Create Not Specified!");
			return false;
		case ENOMEM:
			Debug::outDebug("SharedMemorySegment :: Failed to Get Segment -- Out of Memory!");
			return false;
		case ENOSPC:
			Debug::outDebug("SharedMemorySegment :: Failed to Get Segment -- Too Many Shared Memory Segments!");
			return false;
		case EPERM:
			Debug::outDebug("SharedMemorySegment :: Failed to Get Segment -- No Privileges!");
			return false;
		default:
			Debug::outDebug("SharedMemorySegment :: Failed to Get Segment -- Unknown Error!");
			return false;
		}
	}
	m_Data = shmat(m_SHMid, NULL, 0);
	if (m_Data == (char *)(-1)) {
		Debug::outDebug("SharedMemorySegment :: Failed to Attach to Shared Memory Segment -- Trying Again!");
		closeSHM();
	}

	memset(m_Data, 0, m_DataSize);
	updateSegmentInfo();

	return true;
}

void SharedMemorySegment::closeSHM() {
	if (m_Data) {
		shmdt((void *) m_Data);
		m_Data = NULL;
	}
	if (m_HardClose)
		shmctl(m_SHMid, IPC_RMID, NULL);
	m_SHMid = -1;
	m_SHMkey = 0;
}

void SharedMemorySegment::startMonitoringSegmentForChanges() {
	if (!m_SHMkey) {
		Debug::outError("SharedMemorySegment :: Try Creating a Segment First!");
		return;
	}
}

bool SharedMemorySegment::checkEventsWaiting() {
	// get last change time
	shmid_ds ds;
	if (shmctl(m_SHMid, IPC_STAT, &ds) == -1) {
		Debug::outError("SharedMemorySegment :: Failed to get Info!: %s", strerror(errno));
		return false;
	}

	if ( (ds.shm_ctime != m_ShmDS.shm_ctime) || (m_CurContents.compareWithString((char *) m_Data) != 0) ) {
		onSharedMemorySegmentChange(m_Data);
		updateSegmentInfo();
		return true;
	}

	return false;
}

void SharedMemorySegment::setShmPollDelay(int Delay) {
	m_Delay = Delay;
	Debug::outDebug("SharedMemorySegment :: Set CPU Poll Delay [%d]", Delay);
}

int SharedMemorySegment::getShmPollDelay() {
	return m_Delay;
}

void * SharedMemorySegment::getSharedMemorySegment() {
	return m_Data;
}

void SharedMemorySegment::updateSegmentInfo() {
	if (shmctl(m_SHMid, IPC_STAT, &m_ShmDS) == -1) {
		Debug::outError("SharedMemorySegment :: Failed to get Info!: %s", strerror(errno));
	}

	m_CurContents = (char *) m_Data;
}

bool SharedMemorySegment::signalShmChange() {
	if (shmctl(m_SHMid, IPC_SET, &m_ShmDS) == -1) {
		printf("SharedMemorySegment :: Failed to SET: %s\n\n", strerror(errno));
		return false;
	}

	return true;
}

bool SharedMemorySegment::setShmDataString(const char * data) {
	if (!m_Data)
		return false;

	strcpy((char *) m_Data, data);

	return signalShmChange();
}

