  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//
// GizmoPluginPowerMate.cpp
//
//	GizmoPluginPowerMate class
//
///////////////////////////////////////
  ///////////////////////////////////////
  
/*

Copyright (c) 2005, Tim Burrell
All rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "GizmoPluginPowerMate.hpp"
#include "libGizmo/Debug.hpp"
#include "libGizmo/UtilTime.hpp"
#include <linux/input.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>

////////////////////////////////////////////////////////////////////////////
// Statics / Type Defs
///////////////////////////////////////

#define DEVICE_OPENMODE		O_RDWR
#define NUM_VALID_PREFIXES	2
#define NUM_EVENT_DEVICES	16
#define IBUF_SIZE		32
#define NUM_BLINKS		7

static const char * valid_prefix[NUM_VALID_PREFIXES] = {
	"Griffin PowerMate\0",
	"Griffin SoundKnob\0"
};

int GizmoPluginPowerMate::m_TotalFound = 1;

////////////////////////////////////////////////////////////////////////////
// Overrides
///////////////////////////////////////

void GizmoPluginPowerMate::onClosing() {
	Debug::outDebug("GizmoPluginPowerMate[%s] :: onClosing", m_DevicePath.getString());
	if (m_BlackOnClose)
		onPowerMateSetLED(0);
}

void GizmoPluginPowerMate::onInputRelative(int Code, int Value) {
	Debug::outDebug("GizmoPluginPowerMate[%s] :: onPowerMateRotate: [%d] [%d] Btn = [%d]", m_DevicePath.getString(), Code, Value, m_Button);

	// keep track of turn and press separately
	if (m_Button)
		m_Turned = true;
		
	// do sensitivity
	if (m_RotLast != Value) {
		m_RotLast = Value;
		m_RotCount = 0;
	} else {
		if (++m_RotCount < m_RotSensitivity)
			return;
		else
			m_RotCount = 0;
	}
	
	ScriptFuncArg * pArgDeviceID = m_ScriptFuncOnPowerMateRotate.getInArg(0);
	if (!pArgDeviceID) {
		Debug::outDebug("GizmoPluginPowerMate :: Failed to Call OnPowerMateRotate -- Arg 0");
		return;
	}
	ScriptFuncArg * pArgValue = m_ScriptFuncOnPowerMateRotate.getInArg(1);
	if (!pArgValue) {
		Debug::outDebug("GizmoPluginPowerMate :: Failed to Call OnPowerMateRotate -- Arg 1");
		return;
	}
	ScriptFuncArg * pArgButtonPressed = m_ScriptFuncOnPowerMateRotate.getInArg(2);
	if (!pArgButtonPressed) {
		Debug::outDebug("GizmoPluginPowerMate :: Failed to Call OnPowerMateRotate -- Arg 2");
		return;
	}
	ScriptFuncArg * pArgDevicePath = m_ScriptFuncOnPowerMateRotate.getInArg(3);
	if (!pArgDeviceID) {
		Debug::outDebug("GizmoPluginPowerMate :: Failed to Call OnPowerMateRotate -- Arg 3");
		return;
	}
	
	pArgDeviceID->ArgNumber = m_DeviceID;
	pArgValue->ArgNumber = Value;
	pArgButtonPressed->ArgBoolean = m_Button;
	pArgDevicePath->ArgString = m_DevicePath;
	
	if (m_pScriptEngine)
		m_pScriptEngine->callScriptFunction(&m_ScriptFuncOnPowerMateRotate);
}

void GizmoPluginPowerMate::onInputKey(int Code, int Value) {
	if ( (m_Button = Value) ) {
		Debug::outDebug("GizmoPluginPowerMate :: onPowerMatePress [%d] [%d]", Code, Value);
		ScriptFuncArg * pArgDeviceID = m_ScriptFuncOnPowerMatePress.getInArg(0);
		ScriptFuncArg * pArgDevicePath = m_ScriptFuncOnPowerMatePress.getInArg(1);
		if (m_pScriptEngine && pArgDeviceID && pArgDevicePath) {
			pArgDeviceID->ArgNumber = m_DeviceID;
			pArgDevicePath->ArgString = m_DevicePath;
			m_pScriptEngine->callScriptFunction(&m_ScriptFuncOnPowerMatePress);
		} else {
			Debug::outDebug("GizmoPluginPowerMate :: Failed to Call Event Function");
		}
	} 
	
	else {
		Debug::outDebug("GizmoPluginPowerMate :: onPowerMateRelease [%d] [%d]", Code, Value);
		ScriptFuncArg * pArgDeviceID = m_ScriptFuncOnPowerMateRelease.getInArg(0);
		ScriptFuncArg * pArgDevicePath = m_ScriptFuncOnPowerMateRelease.getInArg(1);
		if (m_pScriptEngine && pArgDeviceID && pArgDevicePath) {
			pArgDeviceID->ArgNumber = m_DeviceID;
			pArgDevicePath->ArgString = m_DevicePath;
			m_pScriptEngine->callScriptFunction(&m_ScriptFuncOnPowerMateRelease);
		} else {
			Debug::outDebug("GizmoPluginPowerMate :: Failed to Call Event Function");
		}
	
		Debug::outDebug("GizmoPluginPowerMate :: onPowerMateClick [%d] [%d]", Code, Value);
		pArgDeviceID = m_ScriptFuncOnPowerMateClick.getInArg(0);
		pArgDevicePath = m_ScriptFuncOnPowerMateClick.getInArg(1);
		if (m_pScriptEngine && pArgDeviceID && pArgDevicePath) {
			pArgDeviceID->ArgNumber = m_DeviceID;
			pArgDevicePath->ArgString = m_DevicePath;
			m_pScriptEngine->callScriptFunction(&m_ScriptFuncOnPowerMateClick);
		} else {
			Debug::outDebug("GizmoPluginPowerMate :: Failed to Call Event Function");
		}
	}

	m_Turned = false;
}

void GizmoPluginPowerMate::onInputMisc(int code, int val) {
	//Debug::outDebug("GizmoPluginPowerMate :: onPowerMateLEDChange: code = [%d] val = [%d]", code, val);
}

bool GizmoPluginPowerMate::onPluginOpened() {
	Debug::outDebug("GizmoPluginPowerMate :: onPowerMateOpened");
	
	return initPlugin();
}

void GizmoPluginPowerMate::onPowerMateSetLED(int Level) {
	setLED((unsigned char) Level);
}

int GizmoPluginPowerMate::onPowerMateGetLED() {
	return (int) getLEDLevel();
}

void GizmoPluginPowerMate::onPowerMateSetRotateSensitivity(int Level) {
	Debug::outDebug("GizmoPluginPowerMate::onPowerMateSetRotateSensitivity :: Level Changed to [%d]", Level);
	m_RotSensitivity = Level;
	if (m_RotSensitivity < 0)
		m_RotSensitivity = 0;
}

void GizmoPluginPowerMate::onPowerMateSetSleepModePulsing(bool DoPulse) {
	// set sleep mode pulsing
	m_PulseAsleep = DoPulse;
	setLED(getLEDLevel());

	Debug::outDebug("GizmoPluginPowerMate :: Sleep Mode Pulsing Set to [%s]", DoPulse ? "true" : "false");
}

void GizmoPluginPowerMate::onPowerMateSetPulsing(int Speed, int Table) {
	pulseLED(Table, Speed);
}

////////////////////////////////////////////////////////////////////////////
// Thread Callback
///////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// Construction
///////////////////////////////////////

GizmoPluginPowerMate::GizmoPluginPowerMate(ScriptEngineInterface * pScriptEngine, const char * DevicePath) : 
	GizmoDeviceInput(DevicePath), 
	m_ScriptFuncPowerMateGetLED(this),
	m_ScriptFuncPowerMateGetLEDPercent(this),
	m_ScriptFuncPowerMateSetLED(this),
	m_ScriptFuncPowerMateSetLEDPercent(this),
	m_ScriptFuncPowerMateSetRotateSensitivity(this),
	m_ScriptFuncPowerMateSetSleepModePulsing(this),
	m_ScriptFuncPowerMateBlinkLED(this),
	m_ScriptFuncPowerMateSetPulsing(this)
{
	Debug::outDebug("GizmoPluginPowerMate :: PowerMate Input Plugin Loading [%s]...", DevicePath);
	
	m_pScriptEngine = pScriptEngine;
	m_DevicePath = DevicePath;
	m_Turned = false;
	m_Button = 0;
	m_Prefix = 0;
	m_RotSensitivity = 0;
	m_RotLast = -303;
	m_RotCount = 0;
	m_Level = 0;
	m_Blinking = 0;
	m_BlinkWatching = false;
	m_PulseAsleep = 1;
	m_BlackOnClose = true;
}

GizmoPluginPowerMate::~GizmoPluginPowerMate() {
	closePlugin();
	Debug::outDebug("GizmoPluginPowerMate :: PowerMate Input Plugin Unloaded.");
}

////////////////////////////////////////////////////////////////////////////
// Class Body
///////////////////////////////////////

void GizmoPluginPowerMate::closePlugin() {
	Debug::outDebug("GizmoPluginPowerMate :: PowerMate Input Plugin Closing...");
	
	onClosing();
}

bool GizmoPluginPowerMate::initPlugin() {
	Debug::outDebug("    PowerMate Input Plugin v%s [%s]", PACKAGE_VERSION, m_DevicePath.getString());

	if (!m_pScriptEngine) {
		Debug::outError("GizmoPluginPowerMate :: Invalid Script Engine!");
		return false;
	}

	// try to open the device
	bool ret = false;
	if (m_DevicePath == "auto")
		ret = findPowerMate();
	else if (m_DevicePath.beginsWith("Device"))
		ret = findPowerMate(atoi(m_DevicePath.mid(7)));
	else
		ret = openDevice(m_DevicePath);

	if (!ret) {
		Debug::outDebug("        - PowerMate Input Plugin Failed to Initialize!");
		return false;
	}
	
	// init the LED
	setLED((unsigned char)128);
	
	// register script functions
	Debug::outDebug("GizmoPluginPowerMate :: Registering Functions...");
	m_pScriptEngine->registerScriptFunction(&m_ScriptFuncPowerMateGetLED);
	m_pScriptEngine->registerScriptFunction(&m_ScriptFuncPowerMateGetLEDPercent);
	m_pScriptEngine->registerScriptFunction(&m_ScriptFuncPowerMateSetLED);
	m_pScriptEngine->registerScriptFunction(&m_ScriptFuncPowerMateSetLEDPercent);
	m_pScriptEngine->registerScriptFunction(&m_ScriptFuncPowerMateSetRotateSensitivity);
	m_pScriptEngine->registerScriptFunction(&m_ScriptFuncPowerMateSetSleepModePulsing);
	m_pScriptEngine->registerScriptFunction(&m_ScriptFuncPowerMateBlinkLED);
	m_pScriptEngine->registerScriptFunction(&m_ScriptFuncPowerMateSetPulsing);

	Debug::out("\t- Found [%s] at [%s]", m_DeviceName.getString(), m_DevicePath.getString());	
	Debug::outDebug("GizmoPluginPowerMate :: PowerMate Input Plugin Initialized.");
	return true;
}

bool GizmoPluginPowerMate::findPowerMate(int number) {
	if (number < 0)
		number = m_TotalFound++;
	
	if (number >= 0)
		Debug::outDebug("\t- Looking for PowerMate Device Number [%d]", number);
	
	for (int lp = 0; lp < NUM_VALID_PREFIXES; lp ++) {
		if (openDevice(valid_prefix[lp], number))
			return true;
	}

	return false;
}

void GizmoPluginPowerMate::blinkLED() {
	m_LevelPreBlink = getLEDLevel();
	m_Blinking = NUM_BLINKS;
}

bool GizmoPluginPowerMate::checkEventsWaiting() {
	return GizmoDeviceInput::checkEventsWaiting();
}

void GizmoPluginPowerMate::pulseLED(int static_brightness, int pulse_speed, int pulse_table, int pulse_asleep, int pulse_awake, bool ResetBlink) {
	struct input_event ev;
	memset(&ev, 0, sizeof(struct input_event));

	if (ResetBlink)
		m_Blinking = 0;
	
	static_brightness &= 0xFF;

	if (pulse_speed < 0)
		pulse_speed = 0;
	if (pulse_speed > 510)
		pulse_speed = 510;
	if (pulse_table < 0)
		pulse_table = 0;
	if (pulse_table > 2)
		pulse_table = 2;
	pulse_asleep = !!pulse_asleep;
	pulse_awake = !!pulse_awake;

	ev.type = EV_MSC;
	ev.code = MSC_PULSELED;
	ev.value = static_brightness | (pulse_speed << 8) | (pulse_table << 17) | (pulse_asleep << 19) | (pulse_awake << 20);

	if (write(m_FD, &ev, sizeof(struct input_event)) != sizeof(struct input_event))
		Debug::outDebug("GizmoPluginPowerMate :: Failed to Write to Griffin PowerMate [%s]", strerror(errno));
}

void GizmoPluginPowerMate::setLED(unsigned char level, bool ResetBlink) {
	int pulse_speed = 255;
	int pulse_table = 0;
	int pulse_awake = 0;

	pulseLED((int)level, pulse_speed, pulse_table, m_PulseAsleep, pulse_awake, ResetBlink);

	m_Level = level;

	//Debug::outDebug("GizmoPluginPowerMate :: LED Brightness to [%d]", level);
}

void GizmoPluginPowerMate::pulseLED(int pulseTable, int pulseSpeed, bool ResetBlink) {
	int pulse_asleep = 1;
	int pulse_awake = 1;

	pulseLED(0x80, pulseSpeed, pulseTable, pulse_asleep, pulse_awake, ResetBlink);

	Debug::outDebug("GizmoPluginPowerMate :: Pulsing LED [Speed %d @ Table %d]", pulseSpeed, pulseTable);
}

unsigned char GizmoPluginPowerMate::getLEDLevel() {
	if (m_Blinking)
		return (unsigned char) m_LevelPreBlink;
	else
		return m_Level;
}

void GizmoPluginPowerMate::setLED(float percent, bool ResetBlink) {
	setLED( (unsigned char) (255.0f * (percent / 100.0f)), ResetBlink);
}

String GizmoPluginPowerMate::getDeviceName() {
	return m_DeviceName;
}

String GizmoPluginPowerMate::getDeviceNameAlt() {
	return m_DevicePath;
}

