  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//
// ScriptEnginePython.cpp
//
//	ScriptEnginePython class
//
///////////////////////////////////////
  ///////////////////////////////////////
  
/*

Copyright (c) 2005, Tim Burrell
All rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "ScriptEnginePython.hpp"
#include "libGizmo/Debug.hpp"
#include "libGizmo/ScriptFunction.hpp"
#include "libGizmo/FileIO.hpp"
#include "libGizmo/UtilTime.hpp"
#include <errno.h>
#include <string.h>
#include <stdlib.h>

////////////////////////////////////////////////////////////////////////////
// Statics / Type Defs
///////////////////////////////////////

#define SCRIPTPLUGIN_PYTHON_EXTENSION	".py"
#define SCRIPTPLUGIN_PYTHON_PROGNAME	"GizmoDaemon"
#define SCRIPTPLUGIN_ENTRY_FUNCTION	"gizmoMain"
#define SCRIPTPLUGIN_PYTHON_MAX_IN_ARGS	20

int 		ScriptEnginePython::m_NumFunctions = 0;
PyMethodDef * 	ScriptEnginePython::m_PyMethods = NULL;
ScriptFunction * ScriptEnginePython::m_Functions [PYTHON_MAX_FUNCTIONS] = {NULL};
PyObject * 	(* ScriptEnginePython::m_PythonFunctions[PYTHON_MAX_FUNCTIONS])(PyObject *, PyObject *) = {NULL};

////////////////////////////////////////////////////////////////////////////
// Static Functions
///////////////////////////////////////

PyObject * ScriptEnginePython::PythonFunctionRunner(PyObject * self, PyObject * args, int Index) {	
	//Debug::outDebug("ScriptEnginePython :: Python Function Called...");
	
	if ( (Index < 0) || (Index >= m_NumFunctions) ) {
		Debug::outError("ScriptEnginePython :: Invalid Python Function Call");
		return NULL;
	}
	
	// process input arguments
	String ArgString;
	void ** VoidArgs;
	int NumArgs = (m_Functions[Index]->InArgList.getSize() > SCRIPTPLUGIN_PYTHON_MAX_IN_ARGS) ? SCRIPTPLUGIN_PYTHON_MAX_IN_ARGS : m_Functions[Index]->InArgList.getSize();
	if (NumArgs > 0) {
		VoidArgs = new void * [NumArgs];
		memset(VoidArgs, 0, sizeof(void *) * NumArgs);
		for (int lp = 0; lp < NumArgs; lp ++) {
			ScriptFuncArg * pArg = m_Functions[Index]->getInArg(lp);
			if (!pArg)
				continue;
			switch (pArg->Type) {
			case SCRIPT_ARG_NUMBER:
				ArgString += "d";
				VoidArgs[lp] = new double;
				break;
			case SCRIPT_ARG_STRING:
				ArgString += "s";
				VoidArgs[lp] = new char *;
				break;
			case SCRIPT_ARG_BOOLEAN:
				ArgString += "i";
				VoidArgs[lp] = new int;
				break;
			}
		}
	
		char * tStr = new char [ArgString.getLength() + 1];
		strncpy(tStr, ArgString.getString(), ArgString.getLength() + 1);
		int argRet = PyArg_ParseTuple(args, tStr, VoidArgs[0], VoidArgs[1], VoidArgs[2], VoidArgs[3], VoidArgs[4], VoidArgs[5], VoidArgs[6], VoidArgs[7], VoidArgs[8], VoidArgs[9], VoidArgs[10], VoidArgs[11], VoidArgs[12], VoidArgs[13], VoidArgs[14], VoidArgs[15], VoidArgs[16], VoidArgs[17], VoidArgs[18], VoidArgs[19]);
		if (!argRet) {
			Debug::outError("ScriptEnginePython :: Invalid Arguments to Python Function [%s]", m_Functions[Index]->FunctionName.getString());
			printf("\t* ");
			PyErr_Print();
		}
		delete [] tStr;
			
		// put the arguments into the ScriptFunction
		for (int lp = 0; lp < NumArgs; lp ++) {
			ScriptFuncArg * pArg = m_Functions[Index]->getInArg(lp);
			if (!pArg)
				continue;
			switch (pArg->Type) {
			case SCRIPT_ARG_NUMBER:
				pArg->ArgNumber = * (double *) VoidArgs[lp];
				break;
			case SCRIPT_ARG_STRING:
				pArg->ArgString = * (char **) VoidArgs[lp];
				break;
			case SCRIPT_ARG_BOOLEAN:
				pArg->ArgBoolean = (bool) * (int *) VoidArgs[lp];
				break;
			}
		}
		
		// clean up
		for (int lp = 0; lp < NumArgs; lp ++) {
			ScriptFuncArg * pArg = m_Functions[Index]->getInArg(lp);
			if (!pArg)
				continue;
			switch (pArg->Type) {
			case SCRIPT_ARG_NUMBER:
				delete (double *) VoidArgs[lp];
				break;
			case SCRIPT_ARG_STRING:
				delete (char **) VoidArgs[lp];
				break;
			case SCRIPT_ARG_BOOLEAN:
				delete (int *) VoidArgs[lp];
				break;
			}
		}
		delete [] VoidArgs;
	}
	
	// execute the function
	if (!m_Functions[Index]->executeFunction()) {
		Debug::outDebug("ScriptEnginePython :: Error in Python Function");
		return NULL;
	}
	
	// get the return value
	PyObject * ret = NULL;
	if (m_Functions[Index]->OutArgList.getSize() == 0) {
		Py_INCREF(Py_None);
		ret = Py_None;
	} else {
		for (int lp = 0; lp < m_Functions[Index]->OutArgList.getSize(); lp ++) {
			ScriptFuncArg * pArg = m_Functions[Index]->getOutArg(lp);
			switch (pArg->Type) {
			case SCRIPT_ARG_NUMBER:
				ret = PyFloat_FromDouble(pArg->ArgNumber);
				break;			
			case SCRIPT_ARG_STRING:
				ret = PyString_FromString(pArg->ArgString.getString());
				break;
			case SCRIPT_ARG_BOOLEAN:
				ret = PyBool_FromLong((long)pArg->ArgBoolean);
				break;
			default:
				Debug::outWarn(" *** ScriptEnginePython :: Code Shouldn't Reach this Line!");
			}
		}
	}
	
	return ret;
}

/////////////////////////////////////////////////////////////////////////////
// Construction
///////////////////////////////////////

ScriptEnginePython::ScriptEnginePython() {
	Debug::outDebug("ScriptEnginePython :: Python Scripting Plugin Loading...");

	initPyFuncCallbacks();
	
	m_Scripting = false;
	m_Initialized = false;
	m_pPyModule = NULL;
	
	pthread_mutex_init(&m_MutexCallScriptFunc, NULL);
}

ScriptEnginePython::~ScriptEnginePython() {
	closePlugin();
	
	pthread_mutex_destroy(&m_MutexCallScriptFunc);
	
	Debug::outDebug("ScriptEnginePython :: Python Scripting Plugin Unloaded.");
}

////////////////////////////////////////////////////////////////////////////
// Class Body
///////////////////////////////////////

void ScriptEnginePython::closePlugin() {
	Debug::outDebug("ScriptEnginePython :: Closing Python Scripting Plugin...");
	
	// close the python interpretter
	if (m_Initialized) {
		Py_Finalize();
		m_Initialized = false;
	}
	
	Debug::outDebug("ScriptEnginePython :: Python Scripting Plugin Closed.");
}

bool ScriptEnginePython::initPlugin() {
	if (m_Initialized)
		return false;
		
	Debug::outDebug("ScriptEnginePython :: Python Scripting Initializing...");
	
	// build PyMethods
	if (!buildPyMethods())
		return false;
	
	Py_SetProgramName(SCRIPTPLUGIN_PYTHON_PROGNAME);
	Py_Initialize();		
	m_pPyGizmoDaemon = Py_InitModule(SCRIPTPLUGIN_PYTHON_PROGNAME, m_PyMethods);
	m_Initialized = true;
	
	Debug::outDebug("ScriptEnginePython :: Python Scripting Plugin Initialized.");
	
	return true;
}

bool ScriptEnginePython::buildPyMethods() {
	Debug::outDebug("ScriptEnginePython :: Building Python Method List of Size [%d]", m_NumFunctions);
	m_PyMethods = new PyMethodDef [m_NumFunctions + 1];	
	for (int lp = 0; lp < m_NumFunctions; lp ++) {
		char * strName = (char *) malloc(sizeof(char) * m_Functions[lp]->FunctionName.getLength() + 1);
		strcpy(strName, m_Functions[lp]->FunctionName.getString());
	
		char * strDescr = (char *) malloc(sizeof(char) * m_Functions[lp]->FunctionDescription.getLength() + 1);
		strcpy(strDescr, m_Functions[lp]->FunctionDescription.getString());
		
		m_PyMethods[lp].ml_name = strName;
		m_PyMethods[lp].ml_meth = m_PythonFunctions[lp];
		m_PyMethods[lp].ml_flags = METH_VARARGS;
		m_PyMethods[lp].ml_doc = strDescr;				
	}
		
	m_PyMethods[m_NumFunctions].ml_name = NULL;
	m_PyMethods[m_NumFunctions].ml_meth = NULL;
	m_PyMethods[m_NumFunctions].ml_flags = 0;
	m_PyMethods[m_NumFunctions].ml_doc = NULL;
		
	return true;
}

bool ScriptEnginePython::checkEventsWaiting() {
	return true;
}

bool ScriptEnginePython::onPluginOpened() {
	Debug::out("    Gizmo Daemon Python Scripting Plugin v%s", PACKAGE_VERSION);
	
	return true;
}

int ScriptEnginePython::runScript(const char * FileName) {
	if (m_Scripting) {
		Debug::outWarn("ScriptEnginePython :: Script [%s] Already Running, Can't Run [%s]", m_ScriptFile.getString(), FileName);
		return -1;
	}
	
	m_ScriptFile = FileName;	
	m_ScriptFile += SCRIPTPLUGIN_PYTHON_EXTENSION;
	Debug::outDebug("ScriptEnginePython :: Launching Script [%s]", m_ScriptFile.getString());
	runScript();	

	return 0;
}

void ScriptEnginePython::runScript() {
	Debug::outDebug("ScriptEnginePython :: Running Script [%s]", m_ScriptFile.getString());
	m_Scripting = true;

	const int PATHS_TO_CHECK = 5;
	String Paths[PATHS_TO_CHECK];

	Paths[0] = m_ScriptFile;
	// try developer dir
	Paths[1] = "../config/" + m_ScriptFile;
	// try home dir
	char * UserHome = getenv("HOME");
	if (UserHome)
		Paths[3] = UserHome + String("/.gizmo/scripts/") + m_ScriptFile;
	// then try prefix/share/plugins/
	Paths[4] = PACKAGE_PREFIX + String("/share/gizmo/scripts/") + m_ScriptFile;

	// now check them in order
	int lp;
	for (lp = 0; lp < PATHS_TO_CHECK; lp ++) {
		Debug::outDebug("\t- Trying to Load Script [%s]", Paths[lp].getString());
		if (FileIO::doesFileExist(Paths[lp])) {
			Debug::out("\t- Loading Script [%s]", Paths[lp].getString());
			m_ScriptFile = Paths[lp];
			
			// execute the script
			executeScript();
			break;
		}
	}
	
	if (lp == PATHS_TO_CHECK) {
		for (lp = 0; lp < PATHS_TO_CHECK; lp ++)
			Debug::out("\t- Failed to Load Script [%s]", Paths[lp].getString());
	}

	m_Scripting = false;
	Debug::outDebug("ScriptEnginePython :: Done Script [%s]", m_ScriptFile.getString());
}

bool ScriptEnginePython::executeScript() {
	// get the script name and path
	String ScriptName = m_ScriptFile;
	String ScriptPath = "./";
	int lSlashPos = m_ScriptFile.findReverse("/");
	if (lSlashPos >= 1) {
		ScriptName = m_ScriptFile.mid(lSlashPos + 1);
		ScriptPath = m_ScriptFile.mid(0, lSlashPos);
		// add the path to Python's sys.path
		PyObject * pstrSys = PyString_FromString("sys");
		PyObject * pmodSys = PyImport_Import(pstrSys);
		Py_DECREF(pstrSys);
		if (pmodSys != NULL) {
			PyObject * sys_path = PyObject_GetAttrString(pmodSys, "path");
			if (sys_path != NULL) {
				PyObject * pstrDir = PyString_FromString(ScriptPath);
				PyList_Append(sys_path, pstrDir);
				Py_DECREF(pstrDir);
			}
			Py_DECREF(sys_path);
		}
		Py_DECREF(pmodSys);
	}
	
	// get the script name into Python Land
	if (ScriptName.find(".py") == ScriptName.getLength() - 3)
		ScriptName = ScriptName.mid(0, ScriptName.getLength() - 4);
	PyObject * pName = PyString_FromString(ScriptName.getString());
	if (!pName) {
		Debug::outError("ScriptEnginePython :: Python Error During Load Script [%s]", m_ScriptFile.getString());
		printf("\t* ");
		PyErr_Print();
		return false;
	}
	
	// load the script
	m_pPyModule = PyImport_Import(pName);
	Py_DECREF(pName);
	// check for errors
	if (!m_pPyModule) {
		Debug::outError("ScriptEnginePython :: Error Loading Script [%s]", m_ScriptFile.getString());
		printf("\t* ");
		PyErr_Print();
		return false;
	}
	
	// call the main function!
	PyObject * pDict = PyModule_GetDict(m_pPyModule);
	PyObject * pFunc = PyDict_GetItemString(pDict, SCRIPTPLUGIN_ENTRY_FUNCTION);
	if (!pFunc) {
		Debug::outError("ScriptEnginePython :: Error Executing %s [%s]", SCRIPTPLUGIN_ENTRY_FUNCTION, m_ScriptFile.getString());
		if (PyErr_Occurred()) {
			printf("\t* ");
			PyErr_Print();
		} else {
			Debug::outError("ScriptEnginePython :: Python Error Unavailable");
		}
		return false;
	}
	
	if (!PyCallable_Check(pFunc)) {
		Debug::outError("ScriptEnginePython :: %s is not Callable! [%s]", SCRIPTPLUGIN_ENTRY_FUNCTION, m_ScriptFile.getString());
		printf("\t* ");
		PyErr_Print();
		return false;
	}
			
	Debug::outDebug("\t- Executing Script...");
	
	PyObject * pValue = PyObject_CallObject(pFunc, NULL);
	if (pValue != NULL) {
		Py_DECREF(pValue);
		return false;
	} else {
		Py_DECREF(pFunc);
		Py_DECREF(m_pPyModule);
		m_pPyModule = NULL;
		PyErr_Print();
		return true;
	}	
}

bool ScriptEnginePython::registerScriptFunction(ScriptFunction * pFunction) {
	if (!pFunction) {
		Debug::out("ScriptEnginPython :: Tried to Register Invalid Function!");
		return false;
	}
	
	if (pFunction->FunctionName == "") {
		Debug::out("ScriptEnginPython :: Tried to Register Function with No Name!");
		return false;
	}
	
	if (pFunction->FunctionDescription == "") {
		Debug::out("ScriptEnginPython :: Tried to Register Function with No Description!");
		return false;
	}
	
	if (m_NumFunctions >= PYTHON_MAX_FUNCTIONS) {
		Debug::out("ScriptEnginPython :: Tried to Register Too Many Functions! Max is [%d]", PYTHON_MAX_FUNCTIONS);
		return false;
	}
	
	// search for duplicates
	for (int lp = 0; lp < m_NumFunctions; lp ++) {
		ScriptFunction * pFunc = m_Functions[lp];
		if (pFunc->FunctionName == pFunction->FunctionName) {
			Debug::outDebug("ScriptEnginePython :: Tried to Reregister Function [%s]", pFunction->FunctionName.getString());
			return true;
		}
	}
	
	m_Functions[m_NumFunctions ++] = pFunction;
	
	// add dynamically after already running
	// different case than adding prior to python init
	if (m_Initialized) {
		//Debug::outDebug("Len1 = %d", pFunction->FunctionName.getLength());
		char * strName = (char *) malloc(sizeof(char) * (pFunction->FunctionName.getLength() + 1));
		strcpy(strName, pFunction->FunctionName.getString());
		
		//Debug::outDebug("Len2 = %d", pFunction->FunctionDescription.getLength());
		char * strDescr = (char *) malloc(sizeof(char) * (pFunction->FunctionDescription.getLength() + 1));
		strcpy(strDescr, pFunction->FunctionDescription.getString());
		
		PyMethodDef * MethodDef = new PyMethodDef [2];
		
		MethodDef[0].ml_name = strName;
		MethodDef[0].ml_meth = m_PythonFunctions[m_NumFunctions - 1];
		MethodDef[0].ml_flags = METH_VARARGS;
		MethodDef[0].ml_doc = strDescr;
		
		MethodDef[1].ml_name = NULL;
		MethodDef[1].ml_meth = NULL;
		MethodDef[1].ml_flags = 0;
		MethodDef[1].ml_doc = NULL;
		
		PyObject * pNewFunc = PyCFunction_New(MethodDef, NULL);
		PyModule_AddObject(m_pPyGizmoDaemon, strName, pNewFunc);
	}
		
	Debug::outDebug("ScriptEnginePython :: Registered Function [%s]", pFunction->FunctionName.getString());
	
	return true;
}

bool ScriptEnginePython::callScriptFunction(ScriptFunction * pFunction, int RecursionDepth) {
	// must make this thread safe!
	const int TIME_DELAY = 50;
	const int MAX_RECURSION = 1024;
	int timeoutCount = 0;
	switch (pthread_mutex_trylock(&m_MutexCallScriptFunc)) {
	default:
		Debug::outError("ScriptEnginePython :: Mutex Error!");
		return false;
	case 0:
		// success!
		break;
	case EDEADLK:
		Debug::outError("ScriptEnginePython :: Deadlock Detected!");
		pthread_mutex_unlock(&m_MutexCallScriptFunc);
		return false;
	case EBUSY:
		Debug::outDebug("ScriptEnginePython :: callScriptFunction busy!");		
		do {
			timeoutCount ++;
			UtilTime::nanoSleep(10000000);
			if (pthread_mutex_trylock(&m_MutexCallScriptFunc) == 0)
				break;
		} while (timeoutCount < TIME_DELAY);
		
		if (timeoutCount < TIME_DELAY) {
			Debug::outDebug("ScriptEnginePython :: callScriptFunction busy recovery");
		} else {
			Debug::outError("ScriptEnginePython :: callScriptFunction unable to recover from busy!");
			if (RecursionDepth < MAX_RECURSION) {
				pthread_mutex_unlock(&m_MutexCallScriptFunc);
				return callScriptFunction(pFunction, RecursionDepth + 1);
			}
		}
	}
	
	// make sure we're in a stable state
	if ( (!pFunction) || (pFunction->FunctionName.getLength() <= 0) || (!m_pPyModule) ) {
		Debug::outDebug("ScriptEnginePythong :: Failed to Call Script Function [%s]", pFunction->FunctionName.getString());
		pthread_mutex_unlock(&m_MutexCallScriptFunc);
		return false;
	}
				
	char * tStr = new char [pFunction->FunctionName.getLength() + 1];
	strcpy(tStr, pFunction->FunctionName.getString());
	PyObject * pFunc = PyObject_GetAttrString(m_pPyModule, tStr);
	delete [] tStr;
	if ( (!pFunc) || (!PyCallable_Check(pFunc)) ) {
		Debug::out("ScriptEnginePython :: Cannot Find Function [%s]", pFunction->FunctionName.getString());
		if (PyErr_Occurred())
 			PyErr_Print();
	        Py_XDECREF(pFunc);
		pthread_mutex_unlock(&m_MutexCallScriptFunc);
	        return false;
        }
        
       	PyObject * pValue = NULL;
        PyObject * pArgs = PyTuple_New(pFunction->InArgList.getSize());
        for (int lp = 0; lp < pFunction->InArgList.getSize(); lp ++) {
        	ScriptFuncArg * pParam = pFunction->InArgList.getDataFromIndex(lp);
        	if (!pParam)
        		continue;
        	switch (pParam->Type) {
        	case SCRIPT_ARG_NUMBER:
        		pValue = PyFloat_FromDouble(pParam->ArgNumber);
        		break;
		case SCRIPT_ARG_STRING:
        		pValue = PyString_FromString(pParam->ArgString.getString());
        		break;
		case SCRIPT_ARG_BOOLEAN:
			pValue = PyInt_FromLong((long) pParam->ArgBoolean);
        		break;
        	default:
        		Debug::outWarn(" *** ScriptEnginePython :: Code Shouldn't Get Here!");
        	}
        	
        	if (!pValue) {
        		Py_DECREF(pArgs);
        		Debug::outDebug("ScriptEnginePython :: Cannot Convert Parameter [%d]", lp);
			pthread_mutex_unlock(&m_MutexCallScriptFunc);
        		return false;
        	}
        	
        	PyTuple_SetItem(pArgs, lp, pValue);
        }
        
        pValue = PyObject_CallObject(pFunc, pArgs);
        Py_DECREF(pArgs);
        if (!pValue) {
        	Py_DECREF(pFunc);
        	Debug::outDebug("ScriptEnginePython :: Failed to Call Function [%s]", pFunction->FunctionName.getString());
        	PyErr_Print();
		pthread_mutex_unlock(&m_MutexCallScriptFunc);
        	return false;
        }
        
        // function call successful, get return params
        if (pFunction->OutArgList.getSize() > 0) {
        	ScriptFuncArg * pParam = pFunction->OutArgList.getDataFromIndex(0);
        	if (!pParam) {
        		Debug::outDebug("ScriptEnginePython :: Call Script Function [%s], OutArgList Invalid!", pFunction->FunctionName.getString());
        		Py_DECREF(pValue);
        		Py_XDECREF(pFunc);
			pthread_mutex_unlock(&m_MutexCallScriptFunc);
        		return false;
        	}
        	
        	switch (pParam->Type) {
        	case SCRIPT_ARG_NUMBER:
        		pParam->ArgNumber = PyFloat_AsDouble(pValue);
        		break;
		case SCRIPT_ARG_STRING:
        		pParam->ArgString = PyString_AsString(pValue);
        		break;
		case SCRIPT_ARG_BOOLEAN:
			pParam->ArgBoolean = (bool) PyInt_AsLong(pValue);
        		break;
        	}
        }
        
	Py_DECREF(pValue);
	Py_XDECREF(pFunc);
	
	pthread_mutex_unlock(&m_MutexCallScriptFunc);
	return true;
}

int ScriptEnginePython::getDeviceID() {
	return 0;
}

String ScriptEnginePython::getDeviceName() {
	return "ScriptEnginePython";
}

String ScriptEnginePython::getDeviceNameAlt() {
	return "ScriptEnginePython";
}

////////////////////////////////////////////////////////////////////////////
// Script Functions
///////////////////////////////////////

void ScriptEnginePython::initPyFuncCallbacks() {
	m_PythonFunctions[0] = PythonFunctionHandler000;
	m_PythonFunctions[1] = PythonFunctionHandler001;
	m_PythonFunctions[2] = PythonFunctionHandler002;
	m_PythonFunctions[3] = PythonFunctionHandler003;
	m_PythonFunctions[4] = PythonFunctionHandler004;
	m_PythonFunctions[5] = PythonFunctionHandler005;
	m_PythonFunctions[6] = PythonFunctionHandler006;
	m_PythonFunctions[7] = PythonFunctionHandler007;
	m_PythonFunctions[8] = PythonFunctionHandler008;
	m_PythonFunctions[9] = PythonFunctionHandler009;
	m_PythonFunctions[10] = PythonFunctionHandler010;
	m_PythonFunctions[11] = PythonFunctionHandler011;
	m_PythonFunctions[12] = PythonFunctionHandler012;
	m_PythonFunctions[13] = PythonFunctionHandler013;
	m_PythonFunctions[14] = PythonFunctionHandler014;
	m_PythonFunctions[15] = PythonFunctionHandler015;
	m_PythonFunctions[16] = PythonFunctionHandler016;
	m_PythonFunctions[17] = PythonFunctionHandler017;
	m_PythonFunctions[18] = PythonFunctionHandler018;
	m_PythonFunctions[19] = PythonFunctionHandler019;
	m_PythonFunctions[20] = PythonFunctionHandler020;
	m_PythonFunctions[21] = PythonFunctionHandler021;
	m_PythonFunctions[22] = PythonFunctionHandler022;
	m_PythonFunctions[23] = PythonFunctionHandler023;
	m_PythonFunctions[24] = PythonFunctionHandler024;
	m_PythonFunctions[25] = PythonFunctionHandler025;
	m_PythonFunctions[26] = PythonFunctionHandler026;
	m_PythonFunctions[27] = PythonFunctionHandler027;
	m_PythonFunctions[28] = PythonFunctionHandler028;
	m_PythonFunctions[29] = PythonFunctionHandler029;
	m_PythonFunctions[30] = PythonFunctionHandler030;
	m_PythonFunctions[31] = PythonFunctionHandler031;
	m_PythonFunctions[32] = PythonFunctionHandler032;
	m_PythonFunctions[33] = PythonFunctionHandler033;
	m_PythonFunctions[34] = PythonFunctionHandler034;
	m_PythonFunctions[35] = PythonFunctionHandler035;
	m_PythonFunctions[36] = PythonFunctionHandler036;
	m_PythonFunctions[37] = PythonFunctionHandler037;
	m_PythonFunctions[38] = PythonFunctionHandler038;
	m_PythonFunctions[39] = PythonFunctionHandler039;
	m_PythonFunctions[40] = PythonFunctionHandler040;
	m_PythonFunctions[41] = PythonFunctionHandler041;
	m_PythonFunctions[42] = PythonFunctionHandler042;
	m_PythonFunctions[43] = PythonFunctionHandler043;
	m_PythonFunctions[44] = PythonFunctionHandler044;
	m_PythonFunctions[45] = PythonFunctionHandler045;
	m_PythonFunctions[46] = PythonFunctionHandler046;
	m_PythonFunctions[47] = PythonFunctionHandler047;
	m_PythonFunctions[48] = PythonFunctionHandler048;
	m_PythonFunctions[49] = PythonFunctionHandler049;
	m_PythonFunctions[50] = PythonFunctionHandler050;
	m_PythonFunctions[51] = PythonFunctionHandler051;
	m_PythonFunctions[52] = PythonFunctionHandler052;
	m_PythonFunctions[53] = PythonFunctionHandler053;
	m_PythonFunctions[54] = PythonFunctionHandler054;
	m_PythonFunctions[55] = PythonFunctionHandler055;
	m_PythonFunctions[56] = PythonFunctionHandler056;
	m_PythonFunctions[57] = PythonFunctionHandler057;
	m_PythonFunctions[58] = PythonFunctionHandler058;
	m_PythonFunctions[59] = PythonFunctionHandler059;
	m_PythonFunctions[60] = PythonFunctionHandler060;
	m_PythonFunctions[61] = PythonFunctionHandler061;
	m_PythonFunctions[62] = PythonFunctionHandler062;
	m_PythonFunctions[63] = PythonFunctionHandler063;
	m_PythonFunctions[64] = PythonFunctionHandler064;
	m_PythonFunctions[65] = PythonFunctionHandler065;
	m_PythonFunctions[66] = PythonFunctionHandler066;
	m_PythonFunctions[67] = PythonFunctionHandler067;
	m_PythonFunctions[68] = PythonFunctionHandler068;
	m_PythonFunctions[69] = PythonFunctionHandler069;
	m_PythonFunctions[70] = PythonFunctionHandler070;
	m_PythonFunctions[71] = PythonFunctionHandler021;
	m_PythonFunctions[72] = PythonFunctionHandler072;
	m_PythonFunctions[73] = PythonFunctionHandler073;
	m_PythonFunctions[74] = PythonFunctionHandler074;
	m_PythonFunctions[75] = PythonFunctionHandler075;
	m_PythonFunctions[76] = PythonFunctionHandler076;
	m_PythonFunctions[77] = PythonFunctionHandler077;
	m_PythonFunctions[78] = PythonFunctionHandler078;
	m_PythonFunctions[79] = PythonFunctionHandler079;
	m_PythonFunctions[80] = PythonFunctionHandler080;
	m_PythonFunctions[81] = PythonFunctionHandler081;
	m_PythonFunctions[82] = PythonFunctionHandler082;
	m_PythonFunctions[83] = PythonFunctionHandler083;
	m_PythonFunctions[84] = PythonFunctionHandler084;
	m_PythonFunctions[85] = PythonFunctionHandler085;
	m_PythonFunctions[86] = PythonFunctionHandler086;
	m_PythonFunctions[87] = PythonFunctionHandler087;
	m_PythonFunctions[88] = PythonFunctionHandler088;
	m_PythonFunctions[89] = PythonFunctionHandler089;
	m_PythonFunctions[90] = PythonFunctionHandler090;
	m_PythonFunctions[91] = PythonFunctionHandler091;
	m_PythonFunctions[92] = PythonFunctionHandler092;
	m_PythonFunctions[93] = PythonFunctionHandler093;
	m_PythonFunctions[94] = PythonFunctionHandler094;
	m_PythonFunctions[95] = PythonFunctionHandler095;
	m_PythonFunctions[96] = PythonFunctionHandler096;
	m_PythonFunctions[97] = PythonFunctionHandler097;
	m_PythonFunctions[98] = PythonFunctionHandler098;
	m_PythonFunctions[99] = PythonFunctionHandler099;
}

PyObject * ScriptEnginePython::PythonFunctionHandler000(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 0);
}

PyObject * ScriptEnginePython::PythonFunctionHandler001(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 1);
}

PyObject * ScriptEnginePython::PythonFunctionHandler002(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 2);
}

PyObject * ScriptEnginePython::PythonFunctionHandler003(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 3);
}

PyObject * ScriptEnginePython::PythonFunctionHandler004(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 4);
}

PyObject * ScriptEnginePython::PythonFunctionHandler005(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 5);
}

PyObject * ScriptEnginePython::PythonFunctionHandler006(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 6);
}

PyObject * ScriptEnginePython::PythonFunctionHandler007(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 7);
}

PyObject * ScriptEnginePython::PythonFunctionHandler008(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 8);
}

PyObject * ScriptEnginePython::PythonFunctionHandler009(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 9);
}

PyObject * ScriptEnginePython::PythonFunctionHandler010(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 10);
}

PyObject * ScriptEnginePython::PythonFunctionHandler011(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 11);
}

PyObject * ScriptEnginePython::PythonFunctionHandler012(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 12);
}

PyObject * ScriptEnginePython::PythonFunctionHandler013(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 13);
}

PyObject * ScriptEnginePython::PythonFunctionHandler014(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 14);
}

PyObject * ScriptEnginePython::PythonFunctionHandler015(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 15);
}

PyObject * ScriptEnginePython::PythonFunctionHandler016(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 16);
}

PyObject * ScriptEnginePython::PythonFunctionHandler017(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 17);
}

PyObject * ScriptEnginePython::PythonFunctionHandler018(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 18);
}

PyObject * ScriptEnginePython::PythonFunctionHandler019(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 19);
}

PyObject * ScriptEnginePython::PythonFunctionHandler020(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 20);
}

PyObject * ScriptEnginePython::PythonFunctionHandler021(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 21);
}

PyObject * ScriptEnginePython::PythonFunctionHandler022(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 22);
}

PyObject * ScriptEnginePython::PythonFunctionHandler023(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 23);
}

PyObject * ScriptEnginePython::PythonFunctionHandler024(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 24);
}

PyObject * ScriptEnginePython::PythonFunctionHandler025(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 25);
}

PyObject * ScriptEnginePython::PythonFunctionHandler026(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 26);
}

PyObject * ScriptEnginePython::PythonFunctionHandler027(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 27);
}

PyObject * ScriptEnginePython::PythonFunctionHandler028(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 28);
}

PyObject * ScriptEnginePython::PythonFunctionHandler029(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 29);
}

PyObject * ScriptEnginePython::PythonFunctionHandler030(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 30);
}

PyObject * ScriptEnginePython::PythonFunctionHandler031(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 31);
}

PyObject * ScriptEnginePython::PythonFunctionHandler032(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 32);
}

PyObject * ScriptEnginePython::PythonFunctionHandler033(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 33);
}

PyObject * ScriptEnginePython::PythonFunctionHandler034(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 34);
}

PyObject * ScriptEnginePython::PythonFunctionHandler035(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 35);
}

PyObject * ScriptEnginePython::PythonFunctionHandler036(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 36);
}

PyObject * ScriptEnginePython::PythonFunctionHandler037(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 37);
}

PyObject * ScriptEnginePython::PythonFunctionHandler038(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 38);
}

PyObject * ScriptEnginePython::PythonFunctionHandler039(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 39);
}

PyObject * ScriptEnginePython::PythonFunctionHandler040(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 40);
}

PyObject * ScriptEnginePython::PythonFunctionHandler041(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 41);
}

PyObject * ScriptEnginePython::PythonFunctionHandler042(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 42);
}

PyObject * ScriptEnginePython::PythonFunctionHandler043(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 43);
}

PyObject * ScriptEnginePython::PythonFunctionHandler044(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 44);
}

PyObject * ScriptEnginePython::PythonFunctionHandler045(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 45);
}

PyObject * ScriptEnginePython::PythonFunctionHandler046(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 46);
}

PyObject * ScriptEnginePython::PythonFunctionHandler047(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 47);
}

PyObject * ScriptEnginePython::PythonFunctionHandler048(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 48);
}

PyObject * ScriptEnginePython::PythonFunctionHandler049(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 49);
}

PyObject * ScriptEnginePython::PythonFunctionHandler050(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 50);
}

PyObject * ScriptEnginePython::PythonFunctionHandler051(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 51);
}

PyObject * ScriptEnginePython::PythonFunctionHandler052(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 52);
}

PyObject * ScriptEnginePython::PythonFunctionHandler053(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 53);
}

PyObject * ScriptEnginePython::PythonFunctionHandler054(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 54);
}

PyObject * ScriptEnginePython::PythonFunctionHandler055(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 55);
}

PyObject * ScriptEnginePython::PythonFunctionHandler056(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 56);
}

PyObject * ScriptEnginePython::PythonFunctionHandler057(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 57);
}

PyObject * ScriptEnginePython::PythonFunctionHandler058(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 58);
}

PyObject * ScriptEnginePython::PythonFunctionHandler059(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 59);
}

PyObject * ScriptEnginePython::PythonFunctionHandler060(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 60);
}

PyObject * ScriptEnginePython::PythonFunctionHandler061(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 61);
}

PyObject * ScriptEnginePython::PythonFunctionHandler062(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 62);
}

PyObject * ScriptEnginePython::PythonFunctionHandler063(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 63);
}

PyObject * ScriptEnginePython::PythonFunctionHandler064(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 64);
}

PyObject * ScriptEnginePython::PythonFunctionHandler065(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 65);
}

PyObject * ScriptEnginePython::PythonFunctionHandler066(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 66);
}

PyObject * ScriptEnginePython::PythonFunctionHandler067(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 67);
}

PyObject * ScriptEnginePython::PythonFunctionHandler068(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 68);
}

PyObject * ScriptEnginePython::PythonFunctionHandler069(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 69);
}

PyObject * ScriptEnginePython::PythonFunctionHandler070(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 70);
}

PyObject * ScriptEnginePython::PythonFunctionHandler071(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 71);
}

PyObject * ScriptEnginePython::PythonFunctionHandler072(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 72);
}

PyObject * ScriptEnginePython::PythonFunctionHandler073(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 73);
}

PyObject * ScriptEnginePython::PythonFunctionHandler074(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 74);
}

PyObject * ScriptEnginePython::PythonFunctionHandler075(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 75);
}

PyObject * ScriptEnginePython::PythonFunctionHandler076(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 76);
}

PyObject * ScriptEnginePython::PythonFunctionHandler077(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 77);
}

PyObject * ScriptEnginePython::PythonFunctionHandler078(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 78);
}

PyObject * ScriptEnginePython::PythonFunctionHandler079(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 79);
}

PyObject * ScriptEnginePython::PythonFunctionHandler080(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 80);
}

PyObject * ScriptEnginePython::PythonFunctionHandler081(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 81);
}

PyObject * ScriptEnginePython::PythonFunctionHandler082(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 82);
}

PyObject * ScriptEnginePython::PythonFunctionHandler083(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 83);
}

PyObject * ScriptEnginePython::PythonFunctionHandler084(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 84);
}

PyObject * ScriptEnginePython::PythonFunctionHandler085(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 85);
}

PyObject * ScriptEnginePython::PythonFunctionHandler086(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 86);
}

PyObject * ScriptEnginePython::PythonFunctionHandler087(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 87);
}

PyObject * ScriptEnginePython::PythonFunctionHandler088(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 88);
}

PyObject * ScriptEnginePython::PythonFunctionHandler089(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 89);
}

PyObject * ScriptEnginePython::PythonFunctionHandler090(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 90);
}

PyObject * ScriptEnginePython::PythonFunctionHandler091(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 91);
}

PyObject * ScriptEnginePython::PythonFunctionHandler092(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 92);
}

PyObject * ScriptEnginePython::PythonFunctionHandler093(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 93);
}

PyObject * ScriptEnginePython::PythonFunctionHandler094(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 94);
}

PyObject * ScriptEnginePython::PythonFunctionHandler095(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 95);
}

PyObject * ScriptEnginePython::PythonFunctionHandler096(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 96);
}

PyObject * ScriptEnginePython::PythonFunctionHandler097(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 97);
}

PyObject * ScriptEnginePython::PythonFunctionHandler098(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 98);
}

PyObject * ScriptEnginePython::PythonFunctionHandler099(PyObject * self, PyObject * args) {
	return PythonFunctionRunner(self, args, 99);
}
